"""
Test for validating the formatting of the json files
"""

import glob
import os
import pytest

from basis_set_exchange import api, validator, fileio
from .common_testvars import all_files, data_dir


@pytest.mark.parametrize('file_path', all_files[0])
def test_valid_meta(file_path):
    full_path = os.path.join(data_dir, file_path)
    validator.validate_file('metadata', full_path)


@pytest.mark.parametrize('file_path', all_files[1])
def test_valid_table(file_path):
    full_path = os.path.join(data_dir, file_path)
    validator.validate_file('table', full_path)


@pytest.mark.parametrize('file_path', all_files[2])
def test_valid_element(file_path):
    full_path = os.path.join(data_dir, file_path)
    validator.validate_file('element', full_path)


@pytest.mark.parametrize('file_path', all_files[3])
def test_valid_component(file_path):
    full_path = os.path.join(data_dir, file_path)
    validator.validate_file('component', full_path)


def test_valid_data_dir():
    validator.validate_data_dir(data_dir)



'''
Comparison of basis data against authoritative sources
'''

from ..api import get_basis
from ..misc import compact_elements
from .readers import read_formatted_basis
from .compare import shells_difference, potentials_difference
from .. import manip


def _print_list(lst):
    '''
    Prints a list from a comparison
    '''

    a = compact_elements(lst)
    if a is not None:
        return a
    else:
        return ""


def basis_comparison_report(bs1, bs2, uncontract_general=False):
    '''
    Compares two basis set dictionaries and prints a report about their differences
    '''

    all_bs1 = list(bs1['basis_set_elements'].keys())

    if uncontract_general:
        bs1 = manip.uncontract_general(bs1)
        bs2 = manip.uncontract_general(bs2)

    not_in_bs1 = []  # Found in bs2, not in bs1
    not_in_bs2 = all_bs1.copy()  # Found in bs1, not in bs2
    no_diff = []  # Elements for which there is no difference
    some_diff = []  # Elements that are different
    big_diff = []  # Elements that are substantially different

    for k, v in bs2['basis_set_elements'].items():
        if k not in all_bs1:
            not_in_bs1.append(k)
            continue

        print()
        print("-------------------------------------")
        print(" Element ", k)
        bs1_el = bs1['basis_set_elements'][k]

        max_rdiff_el = 0.0
        max_rdiff_ecp = 0.0

        if 'element_electron_shells' in v:
            max_rdiff_el = shells_difference(v['element_electron_shells'], bs1_el['element_electron_shells'])
        if 'element_ecp' in v:
            nel1 = v['element_ecp_electrons']
            nel2 = bs1_el['element_ecp_electrons']
            if int(nel1) != int(nel2):
                print('Different number of electrons replaced by ECP ({} vs {})'.format(nel1, nel2))
                max_rdiff_ecp = float('inf')
            else:
                max_rdiff_ecp = potentials_difference(v['element_ecp'], bs1_el['element_ecp'])
                v['element_ecp'] = bs1_el['element_ecp']

        max_rdiff = max(max_rdiff_el, max_rdiff_ecp)

        # Handle some differences
        if max_rdiff == float('inf'):
            big_diff.append(k)
        elif max_rdiff == 0.0:
            no_diff.append(k)
        else:
            some_diff.append(k)

        not_in_bs2.remove(k)

    print()
    print("     Not in bs1: ", _print_list(not_in_bs1))
    print("     Not in bs2: ", _print_list(not_in_bs2))
    print("  No difference: ", _print_list(no_diff))
    print("Some difference: ", _print_list(some_diff))
    print(" BIG difference: ", _print_list(big_diff))
    print()

    return (len(not_in_bs1) == 0 and len(not_in_bs2) == 0 and len(some_diff) == 0 and len(big_diff) == 0)


def compare_basis_against_file(basis_name,
                               src_filepath,
                               file_type=None,
                               version=None,
                               uncontract_general=False,
                               data_dir=None):
    '''Compare a basis set in the BSE against a reference file'''

    src_data = read_formatted_basis(src_filepath, file_type)
    bse_data = get_basis(basis_name, version=version, data_dir=data_dir)
    return basis_comparison_report(src_data, bse_data, uncontract_general=uncontract_general)


def compare_basis_files(file_path_1, file_path_2, file_type_1=None, file_type_2=None, uncontract_general=False):
    '''Compare two files containing formatted basis sets'''

    bs1 = read_formatted_basis(file_path_1, file_type_1)
    bs2 = read_formatted_basis(file_path_2, file_type_2)
    return basis_comparison_report(bs1, bs2, uncontract_general)


def compare_basis_sets(basis_name_1,
                       basis_name_2,
                       version_1=None,
                       version_2=None,
                       uncontract_general=False,
                       data_dir_1=None,
                       data_dir_2=None):
    '''Compare two files containing formatted basis sets'''

    bs1 = get_basis(basis_name_1, version=version_1, data_dir=data_dir_1)
    bs2 = get_basis(basis_name_2, version=version_2, data_dir=data_dir_2)
    return basis_comparison_report(bs1, bs2, uncontract_general)



from ... import lut
from ..skel import create_skel


def read_gbasis(basis_lines, fname):
    '''Reads gbasis-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the gbasis format does not store all the fields we
       have, so some fields are left blank
    '''

    skipchars = '!#'
    basis_lines = [l for l in basis_lines if l and not l[0] in skipchars]

    bs_data = create_skel('component')

    i = 0
    bs_name = None
    while i < len(basis_lines):
        line = basis_lines[i]
        lsplt = line.split(':')
        elementsym = lsplt[0]

        if bs_name is None:
            bs_name = lsplt[1]
        elif lsplt[1] != bs_name:
            raise RuntimeError("Multiple basis sets in a file")

        element_Z = lut.element_Z_from_sym(elementsym)
        element_Z = str(element_Z)

        if not element_Z in bs_data['basis_set_elements']:
            bs_data['basis_set_elements'][element_Z] = {}

        element_data = bs_data['basis_set_elements'][element_Z]

        if not 'element_electron_shells' in element_data:
            element_data['element_electron_shells'] = []

        i += 1

        max_am = int(basis_lines[i].strip())
        i += 1

        for am in range(0, max_am + 1):
            lsplt = basis_lines[i].split()
            shell_am = lut.amchar_to_int(lsplt[0])
            nprim = int(lsplt[1])
            ngen = int(lsplt[2])

            if shell_am[0] != am:
                raise RuntimeError("AM out of order in gbasis?")

            shell = {
                'shell_function_type': 'gto',
                'shell_harmonic_type': 'spherical',
                'shell_region': 'valence',
                'shell_angular_momentum': shell_am
            }

            exponents = []
            coefficients = []

            i += 1
            for j in range(nprim):
                line = basis_lines[i].replace('D', 'E')
                line = line.replace('d', 'E')
                lsplt = line.split()

                if len(lsplt) != (ngen + 1):
                    raise RuntimeError("Incorrect number of general contractions in gbasis")

                exponents.append(lsplt[0])
                coefficients.append(lsplt[1:])
                i += 1

            shell['shell_exponents'] = exponents

            # We need to transpose the coefficient matrix
            # (we store a matrix with primitives being the column index and
            # general contraction being the row index)
            shell['shell_coefficients'] = list(map(list, zip(*coefficients)))

            element_data['element_electron_shells'].append(shell)

    return bs_data



'''
Converts basis set data to a specified output format
'''

from .. import sort
from .bib import write_bib
from .txt import write_txt
from .bsejson import write_json

_converter_map = {
    'txt': {
        'display': 'Plain Text',
        'extension': '.txt',
        'comment': '',
        'function': write_txt
    },
    'bib': {
        'display': 'BibTeX',
        'extension': '.bib',
        'comment': '%',
        'function': write_bib
    },
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': None,
        'function': write_json
    }
}


def convert_references(ref_data, fmt, header=None):
    '''
    Returns the basis set references as a string representing
    the data in the specified output format
    '''

    # Make fmt case insensitive
    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown reference format "{}"'.format(fmt))

    # Sort the data for all references
    for elref in ref_data:
        for rinfo in elref['reference_info']:
            rdata = rinfo['reference_data']
            rinfo['reference_data'] = [(k, sort.sort_single_reference(v)) for k, v in rdata]

    # Actually do the conversion
    ret_str = _converter_map[fmt]['function'](ref_data)

    if header is not None and fmt != 'json':
        comment_str = _converter_map[fmt]['comment']
        header_str = comment_str + comment_str.join(header.splitlines(True))
        ret_str = header_str + '\n\n' + ret_str

    return ret_str


def get_formats():
    '''
    Returns the available reference formats mapped to display name.

    This is returned as an ordered dictionary, with the most common
    at the top, followed by the rest in alphabetical order
    '''

    return {k: v['display'] for k, v in _converter_map.items()}


def get_format_extension(fmt):
    '''
    Returns the recommended extension for a given format
    '''

    if fmt is None:
        return 'dict'

    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    return _converter_map[fmt]['extension']



'''
Testing of a string containing bibtex entries

This tests by running latex/bibtex on a file containing the
string. Warnings are interpreted as failures (which catches
missing fields)
'''

import os
import shutil
import subprocess

# Other modules should query this variable
available = bool(shutil.which('bibtex')) and bool(shutil.which('latex'))

# Test latex file
test_txt = r"""
\documentclass{article}
\begin{document}
\nocite{*}
\bibliographystyle{plain}
\bibliography{bse_test}
\end{document}
"""


def validate_bibtex(tmp_path, bib_str):
    texfile = os.path.join(tmp_path, 'bse.tex')
    bibfile = os.path.join(tmp_path, 'bse_test.bib')

    with open(texfile, 'w') as tf:
        tf.write(test_txt)

    with open(bibfile, 'w') as bf:
        bf.write(bib_str)

    res = subprocess.check_output(['latex', 'bse'], cwd=tmp_path)

    # Run bibtex. Then check return code and scan the output for "Warning"
    res = subprocess.run(
        ['bibtex', 'bse'], universal_newlines=True, cwd=tmp_path, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)

    if "Warning" in res.stdout:
        raise RuntimeError("Warning found in output. Output is:\n" + res.stdout)



"""
Functions related to validating JSON files (including against schema)
"""

import os
import jsonschema

from . import api
from . import fileio


def _validate_extra_references(bs_data):
    '''Extra checks for references files'''
    pass


def _validate_extra_metadata(bs_data):
    '''Extra checks for metadata files'''

    # Check that family is lowercase
    fam = bs_data['basis_set_family']
    if not fam.islower():
        raise RuntimeError("Family '{}' is not lowercase".format(fam))


def _validate_extra_component(bs_data):
    '''Extra checks for component basis files'''

    assert len(bs_data['basis_set_elements']) > 0

    # Make sure size of the coefficient matrix matches the number of exponents
    for el in bs_data['basis_set_elements'].values():
        if not 'element_electron_shells' in el:
            continue

        for s in el['element_electron_shells']:
            nprim = len(s['shell_exponents'])
            if nprim <= 0:
                raise RuntimeError("Invalid number of primitives: {}".format(nprim))

            for g in s['shell_coefficients']:
                if nprim != len(g):
                    raise RuntimeError("Number of coefficients doesn't match number of primitives ({} vs {}".format(
                        len(g), nprim))

            # If more than one AM is given, that should be the number of
            # general contractions
            nam = len(s['shell_angular_momentum'])
            if nam > 1:
                ngen = len(s['shell_coefficients'])
                if ngen != nam:
                    raise RuntimeError("Number of general contractions doesn't match combined AM ({} vs {}".format(
                        ngen, nam))


def _validate_extra_element(bs_data):
    '''Extra checks for basis metadata files'''

    assert len(bs_data['basis_set_elements']) > 0


def _validate_extra_table(bs_data):
    '''Extra checks for table basis files'''

    assert len(bs_data['basis_set_elements']) > 0


_validate_map = {
    'references': _validate_extra_references,
    'metadata': _validate_extra_metadata,
    'component': _validate_extra_component,
    'element': _validate_extra_element,
    'table': _validate_extra_table
}


def validate_data(file_type, bs_data):
    """
    Validates json basis set data against a schema

    Parameters
    ----------
    file_type : str
        Type of file to read. May be 'component', 'element', 'table', or 'references'
    bs_data:
        Data to be validated

    Raises
    ------
    RuntimeError
        If the file_type is not valid (and/or a schema doesn't exist)
    ValidationError
        If the given file does not pass validation
    FileNotFoundError
        If the file given by file_path doesn't exist
    """

    if file_type not in _validate_map:
        raise RuntimeError("{} is not a valid file_type".format(file_type))

    schema = api.get_schema(file_type)
    jsonschema.validate(bs_data, schema)
    _validate_map[file_type](bs_data)


def validate_file(file_type, file_path):
    """
    Validates a file against a schema

    Parameters
    ----------
    file_type : str
        Type of file to read. May be 'component', 'element', 'table', or 'references'
    file_path:
        Full path to the file to be validated

    Raises
    ------
    RuntimeError
        If the file_type is not valid (and/or a schema doesn't exist)
    ValidationError
        If the given file does not pass validation
    FileNotFoundError
        If the file given by file_path doesn't exist
    """

    file_data = fileio._read_plain_json(file_path, False)
    validate_data(file_type, file_data)


def validate_data_dir(data_dir):
    """
    Validates all files in a data_dir
    """

    all_meta, all_table, all_element, all_component = fileio.get_all_filelist(data_dir)

    for f in all_meta:
        full_path = os.path.join(data_dir, f)
        validate_file('metadata', full_path)
    for f in all_table:
        full_path = os.path.join(data_dir, f)
        validate_file('table', full_path)
    for f in all_element:
        full_path = os.path.join(data_dir, f)
        validate_file('element', full_path)
    for f in all_component:
        full_path = os.path.join(data_dir, f)
        validate_file('component', full_path)



import pytest

def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true",
                      default=False, help="Run thorough, but much slower, tests")

def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        return
    skip_slow = pytest.mark.skip(reason="Skipping slow test without --runslow option")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



from ... import lut
from ..skel import create_skel


def read_nwchem(basis_lines, fname):
    '''Reads NWChem-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the nwchem format does not store all the fields we
       have, so some fields are left blank
    '''

    skipchars = '#'
    basis_lines = [l for l in basis_lines if l and not l[0] in skipchars]

    bs_data = create_skel('component')

    i = 0

    while i < len(basis_lines):
        line = basis_lines[i]

        if line.lower().startswith('basis'):
            i += 1

            # NWChem doesn't seem to really block by element
            # It just has shells labeled with each element symbol

            while i < len(basis_lines) and not basis_lines[i].lower().startswith('end'):
                line = basis_lines[i]
                lsplt = line.split()
                elementsym = lsplt[0]
                shell_am = lut.amchar_to_int(lsplt[1])

                element_Z = lut.element_Z_from_sym(elementsym)
                element_Z = str(element_Z)

                if not element_Z in bs_data['basis_set_elements']:
                    bs_data['basis_set_elements'][element_Z] = {}
                if not 'element_electron_shells' in bs_data['basis_set_elements'][element_Z]:
                    bs_data['basis_set_elements'][element_Z]['element_electron_shells'] = []

                element_data = bs_data['basis_set_elements'][element_Z]

                shell = {
                    'shell_function_type': 'gto',
                    'shell_harmonic_type': 'spherical',
                    'shell_region': 'valence',
                    'shell_angular_momentum': shell_am
                }

                exponents = []
                coefficients = []

                i += 1
                while True:
                    if i >= len(basis_lines) or basis_lines[i][0].isalpha():
                        break

                    line = basis_lines[i].replace('D', 'E')
                    line = line.replace('d', 'E')
                    lsplt = line.split()
                    exponents.append(lsplt[0])
                    coefficients.append(lsplt[1:])
                    i += 1

                shell['shell_exponents'] = exponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                shell['shell_coefficients'] = list(map(list, zip(*coefficients)))

                element_data['element_electron_shells'].append(shell)

        elif line.lower().startswith('ecp'):
            i += 1
            while i < len(basis_lines) and not basis_lines[i].lower().startswith('end'):
                line = basis_lines[i]
                if 'nelec' in line.lower():
                    lsplt = line.split()
                    elementsym = lsplt[0]
                    n_elec = int(lsplt[2])

                    element_Z = lut.element_Z_from_sym(elementsym)
                    element_Z = str(element_Z)

                    if not element_Z in bs_data['basis_set_elements']:
                        bs_data['basis_set_elements'][element_Z] = {}
                    if not 'element_ecp_electrons' in bs_data['basis_set_elements'][element_Z]:
                        bs_data['basis_set_elements'][element_Z]['element_ecp_electrons'] = n_elec

                    i += 1
                    continue

                # Now parsing a shell
                lsplt = line.split()
                elementsym = lsplt[0]

                shell_am = lsplt[1]
                if shell_am.lower() == 'ul':
                    shell_am = -1  # Placeholder - leave for later
                else:
                    shell_am = lut.amchar_to_int(lsplt[1])

                element_Z = lut.element_Z_from_sym(elementsym)
                element_Z = str(element_Z)

                if not element_Z in bs_data['basis_set_elements']:
                    bs_data['basis_set_elements'][element_Z] = {}
                if not 'element_ecp' in bs_data['basis_set_elements'][element_Z]:
                    bs_data['basis_set_elements'][element_Z]['element_ecp'] = []
                element_data = bs_data['basis_set_elements'][element_Z]

                ecp_shell = {'potential_angular_momentum': shell_am, 'potential_ecp_type': 'scalar'}

                rexponents = []
                gexponents = []
                coefficients = []

                i += 1
                while True:
                    if i >= len(basis_lines) or basis_lines[i][0].isalpha():
                        break

                    line = basis_lines[i].replace('D', 'E')
                    line = line.replace('d', 'E')
                    lsplt = line.split()
                    rexponents.append(int(lsplt[0]))
                    gexponents.append(lsplt[1])
                    coefficients.append(lsplt[2:])
                    i += 1

                ecp_shell['potential_r_exponents'] = rexponents
                ecp_shell['potential_gaussian_exponents'] = gexponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                ecp_shell['potential_coefficients'] = list(map(list, zip(*coefficients)))

                element_data['element_ecp'].append(ecp_shell)
        i += 1

        # Fix ecp angular momentum now that everything has been read
        for el, v in bs_data['basis_set_elements'].items():
            if not 'element_ecp' in v:
                continue

            max_ecp_am = -1
            for s in v['element_ecp']:
                if s['potential_angular_momentum'] == -1:
                    continue
                max_ecp_am = max(max_ecp_am, max(s['potential_angular_momentum']))

            for s in v['element_ecp']:
                if s['potential_angular_momentum'] == -1:
                    s['potential_angular_momentum'] = [max_ecp_am + 1]

    return bs_data



"""
Functions for reading and writing the standard JSON-based
basis set format
"""

import codecs
import json
import bz2
import os

from .sort import sort_basis_dict, sort_references_dict


def _read_plain_json(file_path, check_bse):
    """
    Reads a JSON file

    A simple wrapper around json.load that only takes the file name
    If the file does not exist, an exception is thrown.

    If the file does exist, but there is a problem with the JSON formatting,
    the filename is added to the exception information.

    If check_bse is True, this function also make sure the 'molssi_bse_schema' key
    exists in the file.

    Parameters
    ----------
    file_path : str
        Full path to the file to read
    """

    if not os.path.isfile(file_path):
        raise FileNotFoundError('JSON file \'{}\' does not exist, is not '
                                'readable, or is not a file'.format(file_path))

    try:
        if file_path.endswith('.bz2'):
            with bz2.open(file_path, 'rt') as f:
                js = json.load(f)
        else:
            with open(file_path, 'r') as f:
                js = json.load(f)

    except json.decoder.JSONDecodeError as ex:
        raise RuntimeError("File {} contains JSON errors".format(file_path)) from ex

    if check_bse is True:
        # Check for molssi_bse_schema key
        if 'molssi_bse_schema' not in js:
            raise RuntimeError('File {} does not appear to be a BSE JSON file'.format(file_path))

    return js


def _write_plain_json(file_path, js):
    """
    Write information to a JSON file

    This makes sure files are created with the proper encoding and consistent indenting

    Parameters
    ----------
    file_path : str
        Full path to the file to write to. It will be overwritten if it exists
    js : dict
        JSON information to write
    """

    # Disable ascii in the json - this prevents the json writer
    # from escaping everything
    with codecs.open(file_path, 'w', 'utf-8') as f:
        json.dump(js, f, indent=4, ensure_ascii=False)


def read_json_basis(file_path):
    """
    Reads generic basis set information from a JSON file

    After reading, the MolSSI BSE schema information is searched for and if not
    found, an exception is raised.

    Parameters
    ----------
    file_path : str
        Full path to the file to read
    """

    return _read_plain_json(file_path, True)


def read_schema(file_path):
    """
    Reads a JSON schema file

    Parameters
    ----------
    file_path : str
        Full path to the file to read
    """

    return _read_plain_json(file_path, False)


def read_references(file_path):
    """
    Reads a references JSON file

    Parameters
    ----------
    file_path : str
        Full path to the file to read
    """

    return _read_plain_json(file_path, True)


def read_metadata(file_path):
    """
    Reads a file containing the metadata for all the basis sets

    Parameters
    ----------
    file_path : str
        Full path to the file to read
    """

    return _read_plain_json(file_path, False)


def write_json_basis(file_path, bs):
    """
    Write basis set information to a JSON file

    Parameters
    ----------
    file_path : str
        Full path to the file to write to. It will be overwritten if it exists
    bs : dict
        Basis set information to write
    """

    _write_plain_json(file_path, sort_basis_dict(bs))


def write_references(file_path, refs):
    """
    Write reference information to a JSON file

    Parameters
    ----------
    file_path : str
        Full path to the file to write to. It will be overwritten if it exists
    refs : dict
        Reference information to write
    """

    _write_plain_json(file_path, sort_references_dict(refs))


def get_all_filelist(data_dir):
    """
    Returns a tuple containing the following (as lists)

    0. All metadata files
    1. All table basis files
    2. All element basis files
    3. All component basis files

    The paths to all the files are returned as paths relative to data_dir
    """

    all_meta = []
    all_table = []
    all_element = []
    all_component = []

    special = ['METADATA.json', 'REFERENCES.json']

    for root, dirs, files in os.walk(data_dir):
        for basename in files:
            if basename in special:
                continue

            fpath = os.path.join(root, basename)
            fpath = os.path.relpath(fpath, data_dir)

            if basename.endswith('.metadata.json'):
                all_meta.append(fpath)
            elif basename.endswith('.table.json'):
                all_table.append(fpath)
            elif basename.endswith('.element.json'):
                all_element.append(fpath)
            elif basename.endswith('.json'):
                all_component.append(fpath)

    return (all_meta, all_table, all_element, all_component)


def read_notes_file(file_path):
    """
    Returns the contents of a notes file.

    If the notes file does not exist, None is returned
    """

    if not os.path.isfile(file_path):
        return None

    with open(file_path, 'r') as f:
        return f.read()



'''
Functions for comparing basis sets and pieces of basis sets
'''

import operator
import copy
from ..manip import sort_shells, sort_potentials


def _reldiff(a, b):
    """
    Computes the relative difference of two floating-point numbers

    rel = abs(a-b)/min(abs(a), abs(b))

    If a == 0 and b == 0, then 0.0 is returned
    Otherwise if a or b is 0.0, inf is returned.
    """

    a = float(a)
    b = float(b)
    aa = abs(a)
    ba = abs(b)

    if a == 0.0 and b == 0.0:
        return 0.0
    elif a == 0 or b == 0.0:
        return float('inf')

    return abs(a - b) / min(aa, ba)


def _compare_keys(element1, element2, key, compare_func, *args):
    """
    Compares a specific key between two elements of a basis set

    If the key exists in one element but not the other, False is returned.

    If the key exists in neither element, True is returned.

    Parameters
    ----------
    element1 : dict
        Basis info for an element
    element2 : dict
        Basis info for another element
    key : string
        Key to compare in the two elements
    compare_func : function
        Function that returns True if the data under the key is equivalent
        in both elements
    args
        Additional arguments to be passed to compare_Func
    """
    if key in element1 and key in element2:
        if not compare_func(element1[key], element2[key], *args):
            return False
    elif key in element1 or key in element2:
        return False

    return True


def _compare_vector(arr1, arr2, rel_tol):
    """
    Compares two vectors (python lists) for approximate equality.

    Each array contains floats or strings convertible to floats

    This function returns True if both arrays are of the same length
    and each value is within the given relative tolerance.
    """

    length = len(arr1)
    if len(arr2) != length:
        return False

    for i in range(length):
        element_1 = float(arr1[i])
        element_2 = float(arr2[i])

        diff = abs(abs(element_1) - abs(element_2))
        if diff != 0.0:
            rel = _reldiff(element_1, element_2)

            # For a basis set, a relatively coarse comparison
            # should be acceptible
            if rel > rel_tol:
                return False

    return True


def _compare_matrix(mat1, mat2, rel_tol):
    """
    Compares two matrices (nested python lists) for approximate equality.

    Each matrix contains floats or strings convertible to floats

    This function returns True if both matrices are of the same dimensions
    and each value is within the given relative tolerance.
    """

    length = len(mat1)
    if len(mat2) != length:
        return False

    for i in range(length):
        if _compare_vector(mat1[i], mat2[i], rel_tol) is False:
            return False

    return True


def compare_electron_shells(shell1, shell2, compare_meta=False, rel_tol=0.0):
    '''
    Compare two electron shells for approximate equality
    (exponents/coefficients are within a tolerance)

    If compare_meta is True, the metadata is also compared for exact equality. 
    '''

    if shell1['shell_angular_momentum'] != shell2['shell_angular_momentum']:
        return False

    exponents1 = shell1['shell_exponents']
    exponents2 = shell2['shell_exponents']
    coefficients1 = shell1['shell_coefficients']
    coefficients2 = shell2['shell_coefficients']

    if not _compare_vector(exponents1, exponents2, rel_tol):
        return False
    if not _compare_matrix(coefficients1, coefficients2, rel_tol):
        return False
    if compare_meta:
        if shell1['shell_region'] != shell2['shell_region']:
            return False
        if shell1['shell_harmonic_type'] != shell2['shell_harmonic_type']:
            return False
        if shell1['shell_function_type'] != shell2['shell_function_type']:
            return False
        return True
    else:
        return True


def electron_shells_are_subset(subset, superset, compare_meta=False, rel_tol=0.0):
    '''
    Determine if a list of electron shells is a subset of another

    If 'subset' is a subset of the 'superset', True is returned.

    The shells are compared approximately (exponents/coefficients are 
    within a tolerance)

    If compare_meta is True, the metadata is also compared for exact equality. 
    '''
    for item1 in subset:
        for item2 in superset:
            if compare_electron_shells(item1, item2, compare_meta, rel_tol):
                break
        else:
            return False

    return True


def electron_shells_are_equal(shells1, shells2, compare_meta=False, rel_tol=0.0):
    '''
    Determine if a list of electron shells is the same as another

    The shells are compared approximately (exponents/coefficients are 
    within a tolerance)

    If compare_meta is True, the metadata is also compared for exact equality. 
    '''

    # Lists are equal if each is a subset of the other
    # Slow but effective
    return electron_shells_are_subset(shells1, shells2, compare_meta, rel_tol) and electron_shells_are_subset(
        shells2, shells1, compare_meta, rel_tol)


def compare_ecp_pots(potential1, potential2, compare_meta=False, rel_tol=0.0):
    '''
    Compare two ecp potentials for approximate equality
    (exponents/coefficients are within a tolerance)

    If compare_meta is True, the metadata is also compared for exact equality. 
    '''

    if potential1['potential_angular_momentum'] != potential2['potential_angular_momentum']:
        return False

    rexponents1 = potential1['potential_r_exponents']
    rexponents2 = potential2['potential_r_exponents']
    gexponents1 = potential1['potential_gaussian_exponents']
    gexponents2 = potential2['potential_gaussian_exponents']
    coefficients1 = potential1['potential_coefficients']
    coefficients2 = potential2['potential_coefficients']

    # integer comparison
    if rexponents1 != rexponents2:
        return False
    if not _compare_vector(gexponents1, gexponents2, rel_tol):
        return False
    if not _compare_matrix(coefficients1, coefficients2, rel_tol):
        return False
    if compare_meta:
        if potential1['potential_ecp_type'] != potential2['potential_ecp_type']:
            return False
        return True
    else:
        return True


def ecp_pots_are_subset(subset, superset, compare_meta=False, rel_tol=0.0):
    '''
    Determine if a list of ecp potentials is a subset of another

    If 'subset' is a subset of the 'superset', True is returned.

    The potentials are compared approximately (exponents/coefficients are 
    within a tolerance)

    If compare_meta is True, the metadata is also compared for exact equality. 
    '''

    for item1 in subset:
        for item2 in superset:
            if compare_ecp_pots(item1, item2, compare_meta, rel_tol):
                break
        else:
            return False

    return True


def ecp_pots_are_equal(pots1, pots2, compare_meta=False, rel_tol=0.0):
    '''
    Determine if a list of electron shells is the same as another

    The potentials are compared approximately (exponents/coefficients are 
    within a tolerance)

    If compare_meta is True, the metadata is also compared for exact equality. 
    '''

    # Lists are equal if each is a subset of the other
    # Slow but effective
    return ecp_pots_are_subset(pots1, pots2, compare_meta) and ecp_pots_are_subset(pots2, pots1, compare_meta)


def compare_elements(element1,
                     element2,
                     compare_electron_shells_meta=False,
                     compare_ecp_pots_meta=False,
                     compare_meta=False,
                     rel_tol=0.0):
    '''
    Determine if the basis information for two elements is the same as another

    Exponents/coefficients are compared using a tolerance.

    Parameters
    ----------
    element1 : dict
        Basis information for an element
    element2 : dict
        Basis information for another element
    compare_electron_shells_meta : bool
        Compare the metadata of electron shells
    compare_ecp_pots_meta : bool
        Compare the metadata of ECP potentials
    compare_meta : bool
        Compare the overall element metadata
    rel_tol : float
        Maximum relative error that is considered equal
    '''

    if not _compare_keys(element1, element2, 'element_electron_shells', electron_shells_are_equal,
                         compare_electron_shells_meta, rel_tol):
        return False

    if not _compare_keys(element1, element2, 'element_ecp', ecp_pots_are_equal, compare_ecp_pots_meta, rel_tol):
        return False

    if not _compare_keys(element1, element2, 'element_ecp_electrons', operator.eq):
        return False

    if compare_meta:
        if not _compare_keys(element1, element2, 'element_references', operator.eq):
            return False

    return True


def compare_basis(bs1,
                  bs2,
                  compare_electron_shells_meta=False,
                  compare_ecp_pots_meta=False,
                  compare_elements_meta=False,
                  compare_meta=False,
                  rel_tol=0.0):
    '''
    Determine if two basis set dictionaries are the same

    bs1 : dict
        Full basis information
    bs2 : dict
        Full basis information
    compare_electron_shells_meta : bool
        Compare the metadata of electron shells
    compare_ecp_pots_meta : bool
        Compare the metadata of ECP potentials
    compare_elements_meta : bool
        Compare the overall element metadata
    compare_meta: bool
        Compare the metadata for the basis set (name, description, etc)
    rel_tol : float
        Maximum relative error that is considered equal
    '''

    els1 = sorted(list(bs1['basis_set_elements'].keys()))
    els2 = sorted(list(bs2['basis_set_elements'].keys()))
    if not els1 == els2:
        return False

    for el in els1:
        if not compare_elements(
                bs1['basis_set_elements'][el],
                bs2['basis_set_elements'][el],
                compare_electron_shells_meta=compare_electron_shells_meta,
                compare_ecp_pots_meta=compare_ecp_pots_meta,
                compare_meta=compare_elements_meta,
                rel_tol=rel_tol):
            print("Elements failed", el)
            return False
    if compare_meta:
        for k in [
                'basis_set_name', 'basis_set_family', 'basis_set_description', 'basis_set_revision_description',
                'basis_set_role', 'basis_set_auxiliaries'
        ]:
            if not _compare_keys(bs1, bs2, k, operator.eq):
                return False
    return True


def shells_difference(s1, s2):
    """
    Computes and prints the differences between two lists of shells

    If the shells contain a different number primitives,
    or the lists are of different length, inf is returned.
    Otherwise, the maximum relative difference is returned.
    """

    max_rdiff = 0.0
    nsh = len(s1)
    if len(s2) != nsh:
        print("Different number of shells: {} vs {}".format(len(s1), len(s2)))
        return float('inf')

    shells1 = sort_shells(s1)
    shells2 = sort_shells(s2)

    for n in range(nsh):
        sh1 = shells1[n]
        sh2 = shells2[n]

        if sh1['shell_angular_momentum'] != sh2['shell_angular_momentum']:
            print("Different angular momentum for shell {}".format(n))
            return float('inf')

        nprim = len(sh1['shell_exponents'])
        if len(sh2['shell_exponents']) != nprim:
            print("Different number of primitives for shell {}".format(n))
            return float('inf')

        ngen = len(sh1['shell_coefficients'])
        if len(sh2['shell_coefficients']) != ngen:
            print("Different number of general contractions for shell {}".format(n))
            return float('inf')

        for p in range(nprim):
            e1 = sh1['shell_exponents'][p]
            e2 = sh2['shell_exponents'][p]
            r = _reldiff(e1, e2)
            if r > 0.0:
                print("   Exponent {:3}: {:20} {:20} -> {:16.8e}".format(p, e1, e2, r))
            max_rdiff = max(max_rdiff, r)

            for g in range(ngen):
                c1 = sh1['shell_coefficients'][g][p]
                c2 = sh2['shell_coefficients'][g][p]
                r = _reldiff(c1, c2)
                if r > 0.0:
                    print("Coefficient {:3}: {:20} {:20} -> {:16.8e}".format(p, c1, c2, r))
                max_rdiff = max(max_rdiff, r)

    print()
    print("Max relative difference for these shells: {}".format(max_rdiff))
    return max_rdiff


def potentials_difference(p1, p2):
    """
    Computes and prints the differences between two lists of potentials 

    If the shells contain a different number primitives,
    or the lists are of different length, inf is returned.
    Otherwise, the maximum relative difference is returned.
    """

    max_rdiff = 0.0
    np = len(p1)
    if len(p2) != np:
        print("Different number of potentials")
        return float('inf')

    pots1 = sort_potentials(p1)
    pots2 = sort_potentials(p2)

    for n in range(np):
        pot1 = pots1[n]
        pot2 = pots2[n]

        if pot1['potential_angular_momentum'] != pot2['potential_angular_momentum']:
            print("Different angular momentum for potential {}".format(n))
            return float('inf')

        nprim = len(pot1['potential_gaussian_exponents'])
        if len(pot2['potential_gaussian_exponents']) != nprim:
            print("Different number of primitives for potential {}".format(n))
            return float('inf')

        ngen = len(pot1['potential_coefficients'])
        if len(pot2['potential_coefficients']) != ngen:
            print("Different number of general contractions for potential {}".format(n))
            return float('inf')

        for p in range(nprim):
            e1 = pot1['potential_gaussian_exponents'][p]
            e2 = pot2['potential_gaussian_exponents'][p]
            r = _reldiff(e1, e2)
            if r > 0.0:
                print("   Gaussian Exponent {:3}: {:20} {:20} -> {:16.8e}".format(p, e1, e2, r))
            max_rdiff = max(max_rdiff, r)

            e1 = pot1['potential_r_exponents'][p]
            e2 = pot2['potential_r_exponents'][p]
            r = _reldiff(e1, e2)
            if r > 0.0:
                print("          R Exponent {:3}: {:20} {:20} -> {:16.8e}".format(p, e1, e2, r))
            max_rdiff = max(max_rdiff, r)

            for g in range(ngen):
                c1 = pot1['potential_coefficients'][g][p]
                c2 = pot2['potential_coefficients'][g][p]
                r = _reldiff(c1, c2)
                if r > 0.0:
                    print("         Coefficient {:3}: {:20} {:20} -> {:16.8e}".format(p, c1, c2, r))
                max_rdiff = max(max_rdiff, r)

    print()
    print("Max relative difference for these potentials: {}".format(max_rdiff))
    return max_rdiff


def subtract_electron_shells(s1, s2, rel_tol=0.0):
    """
    Returns the difference between two lists of electron shells (s1 - s2)

    This will remove any shells from s1 that are also in s2, within a tolerance
    """

    diff_shells = []
    for sh1 in s1:
        for sh2 in s2:
            if compare_electron_shells(sh1, sh2, rel_tol=rel_tol):
                break
        else:
            diff_shells.append(copy.deepcopy(sh1))

    return diff_shells



'''
Test for unused data
'''

import os
import basis_set_exchange as bse
from .common_testvars import data_dir, all_component_files, all_element_files, all_table_files, bs_metadata


def test_unused_data():
    '''
    Test for any unused data in the data directory
    '''

    # All elements contained in all component files
    all_component_elements = {}
    for cf in all_component_files:
        cpath = os.path.join(data_dir, cf)
        cdata = bse.fileio.read_json_basis(cpath)
        all_component_elements[cpath] = list(cdata['basis_set_elements'].keys())

    # All elements contained in all element files
    # And all element data as read from the file
    all_element_elements = {}
    all_element_data = {}
    for ef in all_element_files:
        efpath = os.path.join(data_dir, ef)
        efdata = bse.fileio.read_json_basis(efpath)
        all_element_elements[efpath] = list(efdata['basis_set_elements'].keys())
        all_element_data[efpath] = efdata['basis_set_elements']

    # Now go through what is reachable through a table file
    for tfile in all_table_files:
        table_path = os.path.join(data_dir, tfile)
        table_data = bse.fileio.read_json_basis(table_path)

        # What element files are linked to this table file
        el_files = [(k, v['element_entry']) for k, v in table_data['basis_set_elements'].items()]

        # Loop over the element files, and remove the corresponding entry
        # from all_component_elements
        for el, el_file in el_files:
            el_file_path = os.path.join(data_dir, el_file)
            el_file_data = all_element_data[el_file_path]

            for cfile in el_file_data[el]['element_components']:
                cfile_path = os.path.join(data_dir, cfile)
                if el in all_component_elements[cfile_path]:
                    all_component_elements[cfile_path].remove(el)

            # Now remove the corresponding entry from all_element_elements
            if el in all_element_elements[el_file_path]:
                all_element_elements[el_file_path].remove(el)

    # See which ones were unused
    found_unused = False

    # Merge into one big dictionary
    remaining = all_component_elements
    remaining.update(all_element_elements)

    for k, v in remaining.items():
        if len(v) == 0:
            continue

        found_unused = True
        for el in v:
            print("Element {:3} in {} not used".format(el, k))

    if found_unused:
        raise RuntimeError("Found unused data")



from ... import lut
from ..skel import create_skel


def read_turbomole(basis_lines, fname):
    '''Reads turbomole-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the turbomole format does not store all the fields we
       have, so some fields are left blank
    '''

    skipchars = '*#$'
    basis_lines = [l for l in basis_lines if l and not l[0] in skipchars]

    bs_data = create_skel('component')

    i = 0
    while i < len(basis_lines):
        line = basis_lines[i]
        elementsym = line.split()[0]

        element_Z = lut.element_Z_from_sym(elementsym)
        element_Z = str(element_Z)

        if not element_Z in bs_data['basis_set_elements']:
            bs_data['basis_set_elements'][element_Z] = {}

        element_data = bs_data['basis_set_elements'][element_Z]

        if "ecp" in line.lower():
            if not 'element_ecp' in element_data:
                element_data['element_ecp'] = []

            i += 1
            line = basis_lines[i]

            lsplt = line.split('=')
            maxam = int(lsplt[2])
            n_elec = int(lsplt[1].split()[0])

            amlist = [maxam]
            amlist.extend(list(range(0, maxam)))

            i += 1
            for shell_am in amlist:
                shell_am2 = lut.amchar_to_int(basis_lines[i][0])[0]
                if shell_am2 != shell_am:
                    raise RuntimeError("AM not in expected order?")

                i += 1

                ecp_shell = {
                    'potential_ecp_type': 'scalar',
                    'potential_angular_momentum': [shell_am],
                }
                ecp_exponents = []
                ecp_rexponents = []
                ecp_coefficients = []

                while i < len(basis_lines) and basis_lines[i][0].isalpha() is False:
                    lsplt = basis_lines[i].split()
                    ecp_exponents.append(lsplt[2])
                    ecp_rexponents.append(int(lsplt[1]))
                    ecp_coefficients.append(lsplt[0])
                    i += 1

                ecp_shell['potential_r_exponents'] = ecp_rexponents
                ecp_shell['potential_gaussian_exponents'] = ecp_exponents
                ecp_shell['potential_coefficients'] = [ecp_coefficients]
                element_data['element_ecp'].append(ecp_shell)

            element_data['element_ecp_electrons'] = n_elec

        else:
            if not 'element_electron_shells' in element_data:
                element_data['element_electron_shells'] = []

            i += 1
            while i < len(basis_lines) and basis_lines[i][0].isalpha() == False:
                lsplt = basis_lines[i].split()
                shell_am = lut.amchar_to_int(lsplt[1])
                nprim = int(lsplt[0])

                shell = {
                    'shell_function_type': 'gto',
                    'shell_harmonic_type': 'spherical',
                    'shell_region': '',
                    'shell_angular_momentum': shell_am
                }

                exponents = []
                coefficients = []

                i += 1
                for j in range(nprim):
                    line = basis_lines[i].replace('D', 'E')
                    line = line.replace('d', 'E')
                    lsplt = line.split()
                    exponents.append(lsplt[0])
                    coefficients.append(lsplt[1:])
                    i += 1

                shell['shell_exponents'] = exponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                shell['shell_coefficients'] = list(map(list, zip(*coefficients)))

                element_data['element_electron_shells'].append(shell)

    return bs_data



'''
Functions for creating skeletons of dictionaries/JSON files
'''

import copy

_skeletons = {
    "component": {
        "molssi_bse_schema": {
            "schema_type": "component",
            "schema_version": "0.1"
        },
        "basis_set_description": "",
        "basis_set_references": [],
        "basis_set_elements": {}
    },
    "element": {
        "molssi_bse_schema": {
            "schema_type": "element",
            "schema_version": "0.1"
        },
        "basis_set_name": "",
        "basis_set_description": "",
        "basis_set_elements": {}
    },
    "table": {
        "molssi_bse_schema": {
            "schema_type": "table",
            "schema_version": "0.1"
        },
        "basis_set_revision_description": "",
        "basis_set_elements": {}
    },
    "metadata": {
        "molssi_bse_schema": {
            "schema_type": "metadata",
            "schema_version": "0.1"
        },
        "basis_set_name": "",
        "basis_set_family": "",
        "basis_set_description": "",
        "basis_set_role": "",
        "basis_set_auxiliaries": {}
    }
}


def create_skel(role):
    '''
    Create the skeleton of a dictionary or JSON file

    A dictionary is returned that contains the "molssi_bse_schema"
    key and other required keys, depending on the role

    role can be either 'component', 'element', or 'table'
    '''

    role = role.lower()
    if not role in _skeletons:
        raise RuntimeError("Role {} not found. Should be 'component', 'element', 'table', or 'metadata'")

    return copy.deepcopy(_skeletons[role])



"""
Tests BSE curation functions
"""

import os
import tempfile

from basis_set_exchange import curate, api, validator, misc, fileio, compose, memo
from .common_testvars import auth_data_dir

# yapf: disable
# Mapping of filename to refs
_source_data = {'cc-pvdz.1.gbasis.bz2': None,
                'def2-tzvp.1.tm.bz2': 'myref',
                'aug-pcj-1.1.gbs.bz2': ['ref1', 'ref2'],
                'def2-ecp.1.tm.bz2': {'rb': 'ref1', 47: 'ref2', '48': 'ref3', '49-xe': 'ref4'}
}

_expected_files = {'test_subdir/test_aug-pcj-1_ref1_ref2.1.json': 'H-Ar',
                   'test_subdir/test_cc-pvdz_noref.1.json':       'H-Ar,Ca-Kr',
                   'test_subdir/test_def2-ecp_noref.1.json':      'Sr-Pd,Cs-Rn',
                   'test_subdir/test_def2-ecp_ref1.1.json':       'Rb',
                   'test_subdir/test_def2-ecp_ref2.1.json':       'Ag',
                   'test_subdir/test_def2-ecp_ref3.1.json':       'Cd',
                   'test_subdir/test_def2-ecp_ref4.1.json':       'In-Xe',
                   'test_subdir/test_def2-tzvp_myref.1.json':     'H-Rn',
                   'test_subdir/test_def2-tzvp.1.element.json':   'H-Rn',
                   'test_subdir/test_aug-pcj-1.1.element.json':   'H-Ar',
                   'test_subdir/test_def2-ecp.1.element.json':    'Rb-Rn',
                   'test_subdir/test_cc-pvdz.1.element.json':     'H-Ar,Ca-Kr',
                   'test_aug-pcj-1.1.table.json':                 'H-Ar',
                   'test_cc-pvdz.1.table.json':                   'H-Ar,Ca-Kr',
                   'test_def2-ecp.1.table.json':                  'Rb-Rn',
                   'test_def2-tzvp.1.table.json':                 'H-Rn'
}
# yapf: enable


def test_add_basis(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5
    for sf, refs in _source_data.items():
        sf_path = os.path.join(auth_data_dir, sf)

        name = sf.split('.')[0]
        curate.add_basis(sf_path, tmp_path, 'test_subdir', 'test_' + name, 'test_basis_' + name, 'test_family',
                         'orbital', 'Test Basis Description: ' + name, '1', 'Test Basis Revision Description', refs)

    md = api.get_metadata(tmp_path)
    assert len(md) == len(_source_data)

    # Re-read
    for sf in _source_data.keys():
        name = name = sf.split('.')[0]
        name = 'test_basis_' + name
        bse_dict = api.get_basis(name, data_dir=tmp_path)

        assert bse_dict['basis_set_family'] == 'test_family'

        # Compare against the file we created from
        sf_path = os.path.join(auth_data_dir, sf)
        assert curate.compare_basis_against_file(name, sf_path, data_dir=tmp_path) == True

        # Check that all the files exist and contain the right elements
        for fpath, elements in _expected_files.items():
            fpath = os.path.join(tmp_path, fpath)
            assert os.path.isfile(fpath)

            fdata = fileio.read_json_basis(fpath)
            expect_elements = set(misc.expand_elements(elements, True))
            assert set(fdata['basis_set_elements'].keys()) == expect_elements

    # Validate the new data dir
    validator.validate_data_dir(tmp_path)



'''
Main interface to Basis Set Exchange functionality
'''

import os
import textwrap

from . import compose
from . import converters
from . import fileio
from . import manip
from . import memo
from . import notes
from . import refconverters
from . import references
from . import misc
from . import lut
from ._version import get_versions

__version__ = get_versions()['version']


def version():
    '''Obtain the version of the basis set exchange library (as a string)'''
    return __version__


# Determine the path to the data directory that is part
# of this installation
_my_dir = os.path.dirname(os.path.abspath(__file__))
_default_data_dir = os.path.join(_my_dir, 'data')
_default_schema_dir = os.path.join(_my_dir, 'schema')

# Main URL of the project
_main_url = 'https://www.basissetexchange.org'


def _get_basis_metadata(name, data_dir):
    '''Get metadata for a single basis set

    If the basis doesn't exist, an exception is raised
    '''

    # Transform the name into an internal representation
    tr_name = misc.transform_basis_name(name)

    # Get the metadata for all basis sets
    metadata = get_metadata(data_dir)

    if not tr_name in metadata:
        raise KeyError("Basis set {} does not exist".format(name))

    return metadata[tr_name]


def _header_string(basis_dict):
    '''Creates a header with information about a basis set

    Information includes description, revision, etc, but not references
    '''

    tw = textwrap.TextWrapper(initial_indent='', subsequent_indent=' ' * 20)

    header = '-' * 70 + '\n'
    header += ' Basis Set Exchange\n'
    header += ' Version ' + version() + '\n'
    header += ' ' + _main_url + '\n'
    header += '-' * 70 + '\n'
    header += '   Basis set: ' + basis_dict['basis_set_name'] + '\n'
    header += tw.fill(' Description: ' + basis_dict['basis_set_description']) + '\n'
    header += '        Role: ' + basis_dict['basis_set_role'] + '\n'
    header += tw.fill('     Version: {}  ({})'.format(basis_dict['basis_set_version'],
                                                      basis_dict['basis_set_revision_description'])) + '\n'
    header += '-' * 70 + '\n'

    return header


def fix_data_dir(data_dir):
    '''
    If data_dir is None, returns the default data_dir. Otherwise,
    returns the data_dir parameter unmodified
    '''

    return _default_data_dir if data_dir is None else data_dir


def get_basis(name,
              elements=None,
              version=None,
              fmt=None,
              uncontract_general=False,
              uncontract_spdf=False,
              uncontract_segmented=False,
              make_general=False,
              optimize_general=False,
              data_dir=None,
              header=True):
    '''Obtain a basis set

    This is the main function for getting basis set information.
    This function reads in all the basis data and returns it either
    as a string or as a python dictionary.

    Parameters
    ----------
    name : str
        Name of the basis set. This is not case sensitive.
    elements : str or list
        List of elements that you want the basis set for.
        Elements can be specified by Z-number (int or str) or by symbol (str).
        If this argument is a str (ie, '1-3,7-10'), it is expanded into a list.
        Z numbers and symbols (case insensitive) can be used interchangeably
        (see :func:`bse.misc.expand_elements`)

        If an empty string or list is passed, or if None is passed (the default),
        all elements for which the basis set is defined are included.
    version : int or str
        Obtain a specific version of this basis set. By default,
        the latest version is returned.
    fmt: str
        The desired output format of the basis set. By default,
        basis set information is returned as a python dictionary. Otherwise,
        if a format is specified, a string is returned.
        Use :func:`bse.api.get_formats` to programmatically obtain the available
        formats.  The `fmt` argument is not case sensitive.

        Available formats are

            * nwchem
            * gaussian94
            * psi4
            * gamess_us
            * turbomole
            * json

    uncontract_general : bool
        If True, remove general contractions by duplicating the set
        of primitive exponents with each vector of coefficients.
        Primitives with zero coefficient are removed, as are duplicate shells.
    uncontract_spdf : bool
        If True, remove general contractions with combined angular momentum (sp, spd, etc)
        by duplicating the set of primitive exponents with each vector of coefficients.
        Primitives with zero coefficient are removed, as are duplicate shells.
    uncontract_segmented : bool
        If True, remove segmented contractions by duplicating each primitive into new shells.
        Each coefficient is set to 1.0
    make_general : bool
        If True, make the basis set as generally-contracted as possible. There will be one
        shell per angular momentum (for each element)
    optimize_general : bool
        Optimize by removing general contractions that contain uncontracted
        functions (see :func:`bse.manip.optimize_general`)
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.

    Returns
    -------
    str or dict
        The basis set in the desired format. If `fmt` is **None**, this will be a python
        dictionary. Otherwise, it will be a string.
    '''

    data_dir = fix_data_dir(data_dir)
    bs_data = _get_basis_metadata(name, data_dir)

    # If version is not specified, use the latest
    if version is None:
        version = bs_data['latest_version']
    else:
        version = str(version)  # Version may be an int

    # Compose the entire basis set (all elements)
    file_relpath = bs_data['versions'][version]['file_relpath']
    basis_dict = compose.compose_table_basis(file_relpath, data_dir)

    # Handle optional arguments
    if elements is not None:
        # Convert to purely a list of strings that represent integers
        elements = misc.expand_elements(elements, True)

        # Did the user pass an empty string or empty list? If so, include
        # all elements
        if len(elements) != 0:
            bs_elements = basis_dict['basis_set_elements']

            # Are elements part of this basis set?
            for el in elements:
                if not el in bs_elements:
                    elsym = lut.element_sym_from_Z(el)
                    raise KeyError("Element {} (Z={}) not found in basis {}".format(elsym, el, name))

            # Set to only the elements we want
            basis_dict['basis_set_elements'] = {k: v for k, v in bs_elements.items() if k in elements}

    needs_pruning = False
    if optimize_general:
        basis_dict = manip.optimize_general(basis_dict)
        needs_pruning = True

    # uncontract_segmented implies uncontract_general
    if uncontract_segmented:
        basis_dict = manip.uncontract_segmented(basis_dict)
        needs_pruning = True
    elif uncontract_general:
        basis_dict = manip.uncontract_general(basis_dict)
        needs_pruning = True

    if uncontract_spdf:
        basis_dict = manip.uncontract_spdf(basis_dict)
        needs_pruning = True

    if make_general:
        basis_dict = manip.make_general(basis_dict)
        needs_pruning = True

    # Remove dead and duplicate shells
    if needs_pruning:
        basis_dict = manip.prune_basis(basis_dict)

    # If fmt is not specified, return as a python dict
    if fmt is None:
        return basis_dict

    if header:
        header_str = _header_string(basis_dict)
    else:
        header_str = None

    return converters.convert_basis(basis_dict, fmt, header_str)


def lookup_basis_by_role(primary_basis, role, data_dir=None):
    '''Lookup the name of an auxiliary basis set given a primary basis set and role

    Parameters
    ----------
    primary_basis : str
        The primary (orbital) basis set that we want the auxiliary
        basis set for. This is not case sensitive.
    role: str
        Desired role/type of auxiliary basis set.
        Use :func:`bse.api.get_roles` to programmatically obtain the available
        formats.  The `fmt` argument is not case sensitive.

        Available roles are

            * jfit
            * jkfit
            * rifit
            * admmfit

    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.

    Returns
    -------
    str
        The name of the auxiliary basis set for the given primary basis
        and role.
    '''

    data_dir = fix_data_dir(data_dir)

    role = role.lower()

    if not role in get_roles():
        raise RuntimeError("Role {} is not a valid role".format(role))

    bs_data = _get_basis_metadata(primary_basis, data_dir)
    auxdata = bs_data['auxiliaries']

    if not role in auxdata:
        raise RuntimeError("Role {} doesn't exist for {}".format(role, primary_basis))

    return auxdata[role]


@memo.BSEMemoize
def get_metadata(data_dir=None):
    '''Obtain the metadata for all basis sets

    The metadata includes information such as the display name of the basis set,
    its versions, and what elements are included in the basis set

    The data is read from the METADATA.json file in the `data_dir` directory.

    Parameters
    ----------
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.
    '''

    data_dir = fix_data_dir(data_dir)
    metadata_file = os.path.join(data_dir, "METADATA.json")
    return fileio.read_metadata(metadata_file)


@memo.BSEMemoize
def get_reference_data(data_dir=None):
    '''Obtain information for all stored references

    This is a nested dictionary with all the data for all the references

    The reference data is read from the REFERENCES.json file in the given
    `data_dir` directory.
    '''

    data_dir = fix_data_dir(data_dir)
    reffile_path = os.path.join(data_dir, 'REFERENCES.json')

    return fileio.read_references(reffile_path)


def get_all_basis_names(data_dir=None):
    '''Obtain a list of all basis set names

    The returned list is the internal representation of the basis set name.

    The data is read from the METADATA.json file in the data directory.

    Parameters
    ----------
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.
    '''

    return sorted(list(get_metadata(data_dir).keys()))


def get_references(basis_name, elements=None, version=None, fmt=None, data_dir=None):
    '''Get the references/citations for a basis set

    Parameters
    ----------
    basis_name : str
        Name of the basis set. This is not case sensitive.
    elements : list
        List of element numbers that you want the basis set for. By default,
        all elements for which the basis set is defined are included.
    version : int
        Obtain a specific version of this basis set. By default,
        the latest version is returned.
    fmt: str
        The desired output format of the basis set references. By default,
        basis set information is returned as a list of dictionaries. Use
        get_reference_formats() to programmatically obtain the available formats.
        The `fmt` argument is not case sensitive.

        Available reference formats are

            * bib
            * tex
            * json

    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.

    Returns
    -------
    str or dict
        The references for the given basis set in the desired format. If `fmt` is **None**, this will be a python
        dictionary. Otherwise, it will be a string.
    '''

    data_dir = fix_data_dir(data_dir)
    basis_dict = get_basis(basis_name, elements=elements, version=version, data_dir=data_dir)

    all_ref_data = get_reference_data(data_dir)
    ref_data = references.compact_references(basis_dict, all_ref_data)

    if fmt is None:
        return ref_data

    return refconverters.convert_references(ref_data, fmt)


def get_basis_family(basis_name, data_dir=None):
    '''Lookup a family by a basis set name
    '''

    data_dir = fix_data_dir(data_dir)
    bs_data = _get_basis_metadata(basis_name, data_dir)
    return bs_data['family']


@memo.BSEMemoize
def get_families(data_dir=None):
    '''Return a list of all basis set families'''
    data_dir = fix_data_dir(data_dir)
    metadata = get_metadata(data_dir)

    families = set()
    for v in metadata.values():
        families.add(v['family'])

    return sorted(list(families))


def filter_basis_sets(substr=None, family=None, role=None, data_dir=None):
    '''Filter basis sets by some criteria

    All parameters are ANDed together and are not case sensitive.

    Parameters
    ----------
    substr : str
        Substring to search for in the basis set name
    family : str
        Family the basis set belongs to
    role : str
        Role of the basis set
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.

    Returns
    -------
    dict
        Basis set metadata that matches the search criteria
    '''

    data_dir = fix_data_dir(data_dir)
    metadata = get_metadata(data_dir)

    # family and role are required to be lowercase (via schema and validation functions)

    if family:
        family = family.lower()
        if not family in get_families(data_dir):
            raise RuntimeError("Family '{}' is not a valid family".format(family))
        metadata = {k: v for k, v in metadata.items() if v['family'] == family}
    if role:
        role = role.lower()
        if not role in get_roles():
            raise RuntimeError("Role '{}' is not a valid role".format(role))
        metadata = {k: v for k, v in metadata.items() if v['role'] == role}
    if substr:
        substr = substr.lower()
        metadata = {k: v for k, v in metadata.items() if substr in k or substr in v['display_name']}

    return metadata


@memo.BSEMemoize
def _family_notes_path(family, data_dir):
    '''Form a path to the notes for a family'''

    data_dir = fix_data_dir(data_dir)

    family = family.lower()
    if not family in get_families(data_dir):
        raise RuntimeError("Family '{}' does not exist".format(family))

    file_name = 'NOTES.' + family.lower()
    file_path = os.path.join(data_dir, file_name)
    return file_path


@memo.BSEMemoize
def _basis_notes_path(name, data_dir):
    '''Form a path to the notes for a basis set'''

    data_dir = fix_data_dir(data_dir)
    bs_data = _get_basis_metadata(name, data_dir)

    # the notes file is the same as the base file name, with a .notes extension
    latest_ver = bs_data['latest_version']
    bs_ver_data = bs_data['versions'][latest_ver]
    rel_path = bs_ver_data['file_relpath']
    filebase = os.path.splitext(rel_path)[0]  # remove .json
    filebase = os.path.splitext(filebase)[0]  # remove .tablejson
    filebase = os.path.splitext(filebase)[0]  # remove .[version]
    file_path = os.path.join(data_dir, filebase + '.notes')
    return file_path


@memo.BSEMemoize
def get_family_notes(family, data_dir=None):
    '''Return a string representing the notes about a basis set family

    If the notes are not found, an empty string is returned
    '''

    file_path = _family_notes_path(family, data_dir)
    notes_str = fileio.read_notes_file(file_path)

    if notes_str is None:
        notes_str = ""

    ref_data = get_reference_data(data_dir)
    return notes.process_notes(notes_str, ref_data)


@memo.BSEMemoize
def has_family_notes(family, data_dir=None):
    '''Check if notes exist for a given family

    Returns True if they exist, false otherwise
    '''

    file_path = _family_notes_path(family, data_dir)
    return os.path.isfile(file_path)


@memo.BSEMemoize
def get_basis_notes(name, data_dir=None):
    '''Return a string representing the notes about a specific basis set

    If the notes are not found, an empty string is returned
    '''

    file_path = _basis_notes_path(name, data_dir)
    notes_str = fileio.read_notes_file(file_path)

    if notes_str is None:
        return ""

    ref_data = get_reference_data(data_dir)
    return notes.process_notes(notes_str, ref_data)


@memo.BSEMemoize
def has_basis_notes(family, data_dir=None):
    '''Check if notes exist for a given basis set

    Returns True if they exist, false otherwise
    '''

    file_path = _basis_notes_path(family, data_dir)
    return os.path.isfile(file_path)


def get_schema(schema_type):
    '''Get a schema that can validate BSE JSON files

       The schema_type represents the type of BSE JSON file to be validated,
       and can be 'component', 'element', 'table', 'metadata', or 'references'.
    '''

    schema_file = "{}-schema.json".format(schema_type)
    file_path = os.path.join(_default_schema_dir, schema_file)

    if not os.path.isfile(file_path):
        raise RuntimeError('Schema file \'{}\' does not exist, is not readable, or is not a file'.format(file_path))

    return fileio.read_schema(file_path)


def get_formats(function_types=None):
    '''Return information about the basis set formats available

    The returned data is a map of format to display name. The format
    can be passed as the fmt argument to :func:`get_basis()`

    If a list is specified for function_types, only those formats
    supporting the given function types will be returned.
    '''

    return converters.get_formats(function_types)


def get_reference_formats():
    '''Return information about the reference/citation formats available

    The returned data is a map of format to display name. The format
    can be passed as the fmt argument to :func:`get_references`
    '''
    return refconverters.get_formats()


def get_roles():
    '''Return information about the available basis set roles available

    The returned data is a map of role to display name. The format
    can be passed as the role argument to fmt argument to :func:`lookup_basis_by_role`
    '''

    # yapf: disable
    return { 'orbital': 'Orbital basis',
             'jfit': 'J-fitting',
             'jkfit': 'JK-fitting',
             'rifit': 'RI-fitting',
             'admmfit': 'Auxiliary-Density Matrix Method Fitting',
             'dftxfit': 'DFT Exchange Fitting',
             'dftcfit': 'DFT Correlation Fitting'
            }
    # yapf: enable


def get_data_dir():
    '''Get the default data directory of this installation'''
    return _default_data_dir



"""
Tests for lookups of elemental data
"""

import pytest
from basis_set_exchange import lut

_all_elements = list(range(1, 119))
_all_elements_str = [str(x) for x in _all_elements]
_all_elements.extend(_all_elements_str)
_all_am = list(range(1, 13))


@pytest.mark.parametrize('Z', _all_elements)
def test_element_data(Z):
    # Cycle through the elements and check that
    # the info returned from the functions is consistent
    data = lut.element_data_from_Z(Z)
    assert data[1] == int(Z)

    assert data == lut.element_data_from_sym(data[0])
    assert data == lut.element_data_from_name(data[2])

    assert data[0] == lut.element_sym_from_Z(Z)
    assert data[2] == lut.element_name_from_Z(Z)

    nsym = lut.element_sym_from_Z(Z, True)
    nname = lut.element_name_from_Z(Z, True)

    assert int(Z) == lut.element_Z_from_sym(nsym.upper())

    # Check capitalization
    assert nsym[0] == data[0][0].upper()
    assert nname[0] == data[2][0].upper()


@pytest.mark.parametrize('am', _all_am)
def test_amchar(am):
    # Check that converting am characters, etc, is consistent
    s = lut.amint_to_char([am])
    assert am == lut.amchar_to_int(s)[0]

    combined = list(range(am + 1))
    s = lut.amint_to_char(combined)
    assert combined == lut.amchar_to_int(s)


def test_amchar_special():
    s = lut.amint_to_char([5, 6, 7])
    assert s == 'hik'
    s = lut.amint_to_char([5, 6, 7], hij=True)
    assert s == 'hij'
    s = lut.amint_to_char([0, 1])
    assert s == 'sp'
    s = lut.amint_to_char([0, 1], use_L=True)
    assert s == 'l'


@pytest.mark.xfail
@pytest.mark.parametrize('Z', [150, '150'])
def test_element_data_fail(Z):
    data = lut.element_data_from_Z(Z)


@pytest.mark.xfail
@pytest.mark.parametrize('am', [[1, -1], [1, 100]])
def test_amint_to_char_fail(am):
    amint_to_char(am)



'''
Functionality for handling basis set and family notes
'''

import textwrap

from . import references


def process_notes(notes, ref_data):
    '''Add reference information to the bottom of a notes file

    `:ref:` tags are removed and the actual reference data is appended
    '''

    ref_keys = ref_data.keys()

    found_refs = set()
    for k in ref_keys:
        if k in notes:
            found_refs.add(k)

    # The block to append
    reference_sec = '\n\n'
    reference_sec += '-------------------------------------\n'
    reference_sec += ' REFERENCES\n'
    reference_sec += '-------------------------------------\n'

    # Add reference data
    if len(found_refs) == 0:
        return notes

    for r in sorted(found_refs):
        rtxt = references.reference_text(ref_data[r])
        reference_sec += r + '\n'
        reference_sec += textwrap.indent(rtxt, ' ' * 4)
        reference_sec += '\n\n'

    return notes + reference_sec



'''
Read a basis set file in a given format
'''

import os
import bz2
from .turbomole import read_turbomole
from .g94 import read_g94
from .nwchem import read_nwchem
from .gbasis import read_gbasis
from .dalton import read_dalton

_type_readers = {
    'turbomole': {
        'display': 'Turbomole',
        'extension': '.tm',
        'reader': read_turbomole
    },
    'gaussian94': {
        'display': 'Gaussian94',
        'extension': '.gbs',
        'reader': read_g94
    },
    'nwchem': {
        'display': 'NWChem',
        'extension': '.nw',
        'reader': read_nwchem
    },
    'dalton': {
        'display': 'Dalton',
        'extension': '.mol',
        'reader': read_dalton
    },
    'gbasis': {
        'display': 'GBasis',
        'extension': '.gbasis',
        'reader': read_gbasis
    }
}


def _fix_uncontracted(basis):
    '''
    Forces the contraction coefficient of uncontracted shells to 1.0
    '''

    for el in basis['basis_set_elements'].values():
        if 'element_electron_shells' not in el:
            continue

        for sh in el['element_electron_shells']:
            if len(sh['shell_coefficients']) == 1 and len(sh['shell_coefficients'][0]) == 1:
                sh['shell_coefficients'][0][0] = '1.0000000'

            # Some uncontracted shells don't have a coefficient
            if len(sh['shell_coefficients']) == 0:
                sh['shell_coefficients'].append(['1.0000000'])

    return basis


def read_formatted_basis(file_path, file_type=None):

    if not os.path.isfile(file_path):
        raise RuntimeError('Basis file path \'{}\' does not exist'.format(file_path))

    if file_type is None:
        for k, v in _type_readers.items():
            ext = v['extension']
            ext_bz2 = ext + '.bz2'
            if file_path.endswith(ext) or file_path.endswith(ext_bz2):
                file_type = k
                break

        else:
            raise RuntimeError("Unable to determine basis set format of '{}'".format(file_path))
    else:
        file_type = file_type.lower()

    if file_type not in _type_readers:
        raise RuntimeError("Unknown file type to read '{}'".format(file_type))

    fname = os.path.basename(file_path)

    # Handle compressed files
    if file_path.endswith('.bz2'):
        with bz2.open(file_path, 'rt') as f:
            basis_lines = [l.strip() for l in f.readlines()]
    else:
        with open(file_path, 'r') as f:
            basis_lines = [l.strip() for l in f.readlines()]

    data = _type_readers[file_type]['reader'](basis_lines, fname)

    # It's debateable if I want to do this
    #return _fix_uncontracted(data)

    return data


def get_reader_formats():
    '''
    Returns the available formats mapped to display name.

    This is returned as an ordered dictionary, with the most common
    at the top, followed by the rest in alphabetical order
    '''

    return {k: v['display'] for k, v in _type_readers.items()}



"""
Tests for the BSE main API
"""

import random

import basis_set_exchange as bse
import pytest

from basis_set_exchange import lut
from .common_testvars import bs_formats, ref_formats


@pytest.mark.parametrize('fmt', bs_formats)
def test_get_format_extensions(fmt):
    """For all basis set formats, get the extension
    """
    bse.converters.get_format_extension(fmt)


@pytest.mark.parametrize('fmt', ref_formats)
def test_get_refformat_extensions(fmt):
    """For all basis set formats, get the extension
    """
    bse.refconverters.get_format_extension(fmt)



'''
Test for duplicate data in a basis set
'''

import os
import pytest
import basis_set_exchange as bse
from basis_set_exchange import curate
from .common_testvars import bs_names_vers, test_data_dir, true_false


def _test_duplicates(bs_dict):
    '''
    Test for any duplicate data in a basis set
    '''

    found_dupe = False
    for el, eldata in bs_dict['basis_set_elements'].items():
        if 'element_electron_shells' in eldata:
            # Quick and dirty
            shells = eldata['element_electron_shells']

            for sh in shells:
                if shells.count(sh) != 1:
                    print("Shell Dupe: element {}".format(el))
                    found_dupe = True
                    break

        if 'element_ecp' in eldata:
            pots = eldata['element_ecp']

            for pot in pots:
                if pots.count(pot) != 1:
                    print("ECP Dupe: element {}".format(el))
                    found_dupe = True
                    break

    assert found_dupe == False


@pytest.mark.parametrize('bs_name,bs_ver', bs_names_vers)
def test_duplicate_data(bs_name, bs_ver):
    '''
    Test for any duplicate data in a basis set
    '''

    bs_dict = bse.get_basis(bs_name, version=bs_ver)

    _test_duplicates(bs_dict)


@pytest.mark.xfail
@pytest.mark.parametrize('filename', ['6-31g-bse-DUPE.nw.bz2', 'def2-ecp-DUPE.nw.bz2'])
def test_duplicate_fail(filename):
    filepath = os.path.join(test_data_dir, filename)

    filedata = curate.read_formatted_basis(filepath, 'nwchem')
    _test_duplicates(filedata)


@pytest.mark.slow
@pytest.mark.parametrize('bs_name,bs_ver', bs_names_vers)
@pytest.mark.parametrize('unc_gen', true_false)
@pytest.mark.parametrize('unc_seg', true_false)
@pytest.mark.parametrize('unc_spdf', true_false)
@pytest.mark.parametrize('make_gen', true_false)
@pytest.mark.parametrize('opt_gen', true_false)
def test_duplicate_slow(bs_name, bs_ver, unc_gen, unc_seg, unc_spdf, opt_gen, make_gen):
    '''
    Test for any duplicate data in a basis set
    '''

    bs_dict = bse.get_basis(bs_name, version=bs_ver,
                            uncontract_general=unc_gen,
                            uncontract_segmented=unc_seg,
                            uncontract_spdf=unc_spdf,
                            make_general=make_gen,
                            optimize_general=opt_gen)

    _test_duplicates(bs_dict)



"""
Tests BSE curation functions
"""

import os
import pytest
import shutil
import bz2

from basis_set_exchange import api, curate, fileio
from .common_testvars import data_dir, test_data_dir


# yapf: disable
@pytest.mark.parametrize('basis1, basis2, element, expected', [
                              ['6-31g', '6-31g*', '6', True],
                              ['6-31g', '6-31g**', '6', True],
                              ['6-31g*', '6-31g*', '6', True],
                              ['6-31g*', '6-31g', '6', False],
                              ['6-31g*', '6-31g', '1', True],
                              ['6-31g**', '6-31g','1', False],
                              ['cc-pvtz', 'aug-cc-pvtz', '13', True]
                         ])
# yapf: enable
def test_electron_subset(basis1, basis2, element, expected):
    el1 = api.get_basis(basis1)['basis_set_elements'][element]
    el2 = api.get_basis(basis2)['basis_set_elements'][element]
    shells1 = el1['element_electron_shells']
    shells2 = el2['element_electron_shells']
    assert curate.electron_shells_are_subset(shells1, shells2, True) == expected


# yapf: disable
@pytest.mark.parametrize('basis1, basis2, element, expected', [
                              ['6-31g', '6-31g', '8', True],
                              ['6-31g', '6-31g*', '8', False],
                              ['6-31g', '6-31g**', '8', False],
                              ['6-31g', '6-31g**', '1', False],
                              ['6-31g', '6-31g*', '1', True],
                              ['cc-pvtz', 'aug-cc-pvtz', '13', False]
                         ])
# yapf: enable
def test_electron_equal(basis1, basis2, element, expected):
    el1 = api.get_basis(basis1)['basis_set_elements'][element]
    el2 = api.get_basis(basis2)['basis_set_elements'][element]
    shells1 = el1['element_electron_shells']
    shells2 = el2['element_electron_shells']
    assert curate.electron_shells_are_equal(shells1, shells2, True) == expected
    assert curate.electron_shells_are_equal(shells2, shells1, True) == expected


# yapf: disable
@pytest.mark.parametrize('basis1, basis2, element, expected', [
                              ['CRENBL', 'CRENBL', '78', True],
                              ['CRENBL', 'CRENBL', '92', True],
                              ['CRENBL', 'CRENBL', '118', True],
                              ['LANL2DZ', 'LANL2DZ', '78', True],
                              ['CRENBL', 'LANL2DZ', '78', False],
                              ['LANL2DZ', 'CRENBL', '78', False]
                         ])
# yapf: enable
def test_ecp_equal(basis1, basis2, element, expected):
    el1 = api.get_basis(basis1)['basis_set_elements'][element]
    el2 = api.get_basis(basis2)['basis_set_elements'][element]
    ecps1 = el1['element_ecp']
    ecps2 = el2['element_ecp']
    assert curate.ecp_pots_are_equal(ecps1, ecps2, True) == expected


# yapf: disable
@pytest.mark.parametrize('basis1, basis2, element, expected', [
                              ['6-31g', '6-31g', '8', True],
                              ['6-31g', '6-31g*', '8', False],
                              ['6-31g', '6-31g**', '8', False],
                              ['6-31g', '6-31g**', '1', False],
                              ['6-31g', '6-31g*', '1', True],
                              ['aug-cc-pvtz', 'aug-cc-pvqz', '15', False],
                              ['aug-cc-pvtz', 'aug-cc-pvdz', '18', False],
                              ['CRENBL', 'CRENBL', '78', True],
                              ['CRENBL', 'CRENBL', '92', True],
                              ['CRENBL', 'CRENBL', '118', True],
                              ['CRENBL', 'LANL2DZ', '78', False],
                              ['LANL2DZ', 'CRENBL', '78', False]
                         ])
# yapf: enable
def test_compare_elements(basis1, basis2, element, expected):
    el1 = api.get_basis(basis1)['basis_set_elements'][element]
    el2 = api.get_basis(basis2)['basis_set_elements'][element]
    assert curate.compare_elements(el1, el2, True, True, True) == expected
    assert curate.compare_elements(el2, el1, True, True, True) == expected


# yapf: disable
@pytest.mark.parametrize('basis, element', [
                              ['6-31g', '8'],
                              ['CRENBL', '3'],
                              ['CRENBL', '92'],
                              ['LANL2DZ', '78']
                         ])
# yapf: enable
def test_printing(basis, element):
    el = api.get_basis(basis)['basis_set_elements'][element]

    shells = el['element_electron_shells']
    curate.print_electron_shell(shells[0])

    if 'element_ecp' in el:
        ecps = el['element_ecp']
        curate.print_ecp_pot(ecps[0])

    curate.print_element(element, el)


# yapf: disable
@pytest.mark.parametrize('file_path', [
                              'dunning/cc-pVDZ_dunning1989a.1.json',
                              'crenb/CRENBL_ross1994a.0.json',
                              'crenb/CRENBL-ECP_ross1994a.0.json'
                         ])
# yapf: enable
def test_print_component_basis(file_path):
    full_path = os.path.join(data_dir, file_path)
    comp = fileio.read_json_basis(full_path)
    curate.print_component_basis(comp)


# yapf: disable
@pytest.mark.parametrize('file_path', [
                              'dunning/cc-pVDZ.1.element.json',
                              'crenb/CRENBL.0.element.json'
                         ])
# yapf: enable
def test_print_elemental_basis(file_path):
    full_path = os.path.join(data_dir, file_path)
    el = fileio.read_json_basis(full_path)
    curate.print_element_basis(el)


# yapf: disable
@pytest.mark.parametrize('file_path', [
                              'cc-pVDZ.1.table.json',
                              'CRENBL.0.table.json'
                         ])
# yapf: enable
def test_print_table_basis(file_path):
    full_path = os.path.join(data_dir, file_path)
    tab = fileio.read_json_basis(full_path)
    curate.print_table_basis(tab)


def test_diff_json_files_same(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5

    filename = 'def2-SV-base_gulde2012a.1.json'
    file1 = os.path.join(data_dir, 'ahlrichs', 'SV', filename)
    tmpfile = os.path.join(tmp_path, filename)
    shutil.copyfile(file1, tmpfile)

    curate.diff_json_files([tmpfile], [tmpfile])

    diff_file = tmpfile + '.diff'
    assert os.path.isfile(diff_file)

    diff_data = fileio.read_json_basis(diff_file)
    assert len(diff_data['basis_set_elements']) == 0


def test_diff_json_files(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5

    filename1 = '6-31G**-full.json.bz2' 
    filename2 = '6-31G-full.json.bz2' 

    file1 = os.path.join(test_data_dir, filename1)
    file2 = os.path.join(test_data_dir, filename2)

    tmpfile1 = os.path.join(tmp_path, filename1)
    tmpfile2 = os.path.join(tmp_path, filename2)

    shutil.copyfile(file1, tmpfile1)
    shutil.copyfile(file2, tmpfile2)

    curate.diff_json_files([tmpfile1], [tmpfile2])
    curate.diff_json_files([tmpfile2], [tmpfile1])

    diff1 = fileio.read_json_basis(tmpfile1 + '.diff')
    diff2 = fileio.read_json_basis(tmpfile2 + '.diff')

    assert len(diff1['basis_set_elements']) == 36
    assert len(diff2['basis_set_elements']) == 0

    reffilename = '6-31G**-polarization.json.bz2' 
    reffile = os.path.join(test_data_dir, reffilename)
    refdata = fileio.read_json_basis(reffile)

    assert curate.compare_basis(diff1, refdata, rel_tol=0.0)



'''
Conversion of references to plain text format
'''

import textwrap
from ..misc import compact_elements
from ..references import reference_text


def write_txt(refs):
    '''Converts references to plain text format
    '''
    full_str = u'** ' + '\n\n'
    for ref in refs:
        full_str += u'{}\n'.format(compact_elements(ref['elements']))

        for ri in ref['reference_info']:
            full_str += u'    ## {}\n'.format(ri['reference_description'])

            refdata = ri['reference_data']

            if len(refdata) == 0:
                full_str += u'    (...no reference...)\n\n'
            for k, r in refdata:
                ref_txt = reference_text(r)
                ref_txt = textwrap.indent(ref_txt, ' ' * 4)
                full_str += u'{}\n\n'.format(ref_txt)

    return full_str



"""
Tests for the BSE main API
"""

import random
import pytest

import basis_set_exchange as bse

pytestmark = pytest.mark.xfail


@pytest.mark.parametrize('basis_name', ['cc-pv0z', 'madeup_name'])
def test_get_basis_fail_name(basis_name):
    """Test get_basis with bad names """
    bse.get_basis(basis_name)


@pytest.mark.parametrize('elements', [100, '100', [1, 100], [1, '100']])
def test_get_basis_fail_elements(elements):
    """Test get_basis with bad elements """

    bse.get_basis('cc-pvdz', elements=elements)


@pytest.mark.parametrize('fmt', ['nwchemm', 'nofmt', 1])
def test_get_basis_fail_fmt(fmt):
    """Test get_basis with bad formats """

    bse.get_basis('cc-pvdz', fmt=fmt)


@pytest.mark.parametrize('version', ['v1', -1, '-1', 100])
def test_get_basis_fail_version(version):
    """Test get_basis with bad versions """

    bse.get_basis('cc-pvdz', version=version)


@pytest.mark.parametrize('fmt', ['nwchemm', 'nofmt', 1])
def test_get_basis_fail_data_dir(tmp_path):
    """Test get_basis with bad data dir """

    tmp_path = str(tmp_path)  # Needed for python 3.5
    bse.get_basis('cc-pvdz', data_dir=tmp_path)


def test_get_basis_fail_data_dir(tmp_path):
    """Test get_basis with bad data dir """

    tmp_path = str(tmp_path)  # Needed for python 3.5
    bse.get_basis('cc-pvdz', data_dir=tmp_path)


def test_get_all_basis_names_fail(tmp_path):
    """Test get_all_basis_names with bad data dir """

    tmp_path = str(tmp_path)  # Needed for python 3.5
    bse.get_all_basis_names(tmp_path)


@pytest.mark.parametrize('basis_name', ['cc-pv0z', 'madeup_name'])
def test_lookup_role_fail_name(basis_name):
    """Test get_basis with bad versions """

    bse.lookup_basis_by_role(basis_name, 'rifit')


@pytest.mark.parametrize('role', ['admmfit', 'madeup_name'])
def test_lookup_role_fail_role(role):
    """Test get_basis with bad versions """

    bse.lookup_basis_by_role('def2-tzvp', role)


# yapf: disable
@pytest.mark.parametrize('substr,family,role', [['def2', 'ahlrichsx', 'jkfit'],
                                                ['qqqqq', None, 'notarole'],
                                                ['6-31', 'poplexx,', 'admmfit']])
# yapf: enable
def test_filter_fail(substr, family, role):
    """Test filtering basis set (returning zero results)"""
    md = bse.filter_basis_sets(substr, family, role)



"""
Tests for reference handling
"""

import json
import os
import pytest

from basis_set_exchange import api, validator, fileio
from .common_testvars import data_dir, all_component_files


@pytest.mark.parametrize('file_path', all_component_files)
def test_filenames(file_path):
    '''
    Test that component filenames end with the correct reference
    '''
    full_path = os.path.join(data_dir, file_path)
    with open(full_path, 'r') as f:
        file_refs = json.load(f)['basis_set_references']
        if len(file_refs) > 0:
            ref_str = '_'.join(file_refs)
        else:
            ref_str = 'noref'
        ref_str = '_' + ref_str

        # Strip off '.0.json', '.1.json', etc
        base_name = os.path.splitext(file_path)[0]
        base_name = os.path.splitext(base_name)[0]

        # Base name should end with the reference string
        print(base_name)
        print(ref_str)
        assert base_name.endswith(ref_str)


def test_valid_reffile():
    '''
    Test to make sure the references file is valid
    '''
    full_path = os.path.join(data_dir, "REFERENCES.json")
    validator.validate_file('references', full_path)



'''
Command line interface for the basis set exchange
'''

import argparse
import argcomplete
from .. import version
from .bsecurate_handlers import bsecurate_cli_handle_subcmd
from .check import cli_check_normalize_args
from .complete import cli_case_insensitive_validator, cli_bsname_completer, cli_readerfmt_completer


def run_bsecurate_cli():
    ################################################################################################
    # NOTE: I am deliberately not using the 'choices' argument in add_argument. I could use it
    # for formats, etc, however I wouldn't want to use it for basis set names. Therefore, I handle
    # all of that manually so that error output is consistent and clean
    ################################################################################################

    ########################################
    # Main global options
    ########################################
    parser = argparse.ArgumentParser(description='Description of your program')
    parser.add_argument('-V', action='version', version='basis_set_exchange ' + version())
    parser.add_argument('-d', '--data-dir', metavar='PATH', help='Override which data directory to use')
    parser.add_argument('-o', '--output', metavar='PATH', help='Output to given file rather than stdout')

    subparsers = parser.add_subparsers(metavar='subcommand', dest='subcmd')
    subparsers.required = True # https://bugs.python.org/issue9253#msg186387

    ########################################
    # Listing of general info and metadata
    ########################################
    # get-reader-formats
    subp = subparsers.add_parser('get-reader-formats', help='A list of file formats that can be read')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the format names')

    # elements-in-files
    subp = subparsers.add_parser('elements-in-files', help='For a list of JSON files, output what elements are in each file')
    subp.add_argument('files', nargs='+', help='List of files to inspect')


    ########################################
    # Manipulating basis set data
    ########################################
    # make-diff
    subp = subparsers.add_parser('make-diff', help='Find/Store the differences between two groups of files')
    subp.add_argument('-l', '--left', nargs='+', required=True, help='Base JSON files')
    subp.add_argument('-r', '--right', nargs='+', required=True, help='JSON files with data to subtract from the base files')


    ########################################
    # Comparing
    ########################################
    # compare-basis-sets
    subp = subparsers.add_parser('compare-basis-sets', help='Compare two basis sets in the data directory')
    subp.add_argument('basis1', help='First basis set to compare').completer = cli_bsname_completer
    subp.add_argument('basis2', help='Second basis set to compare').completer = cli_bsname_completer
    subp.add_argument('--version1', help='Version of the first basis set to compare with. Default is latest')
    subp.add_argument('--version2', help='Version of the second basis set to compare with. Default is latest')
    subp.add_argument('--uncontract-general', action='store_true', help='Remove general contractions before comparing')

    # compare-basis-files
    subp = subparsers.add_parser('compare-basis-files', help='Compare two formatted basis set files')
    subp.add_argument('file1', help='First basis set file to compare')
    subp.add_argument('file2', help='Second basis set file to compare')
    subp.add_argument('--readfmt1', help='Override the file format of file 1').completer = cli_readerfmt_completer
    subp.add_argument('--readfmt2', help='Override the file format of file 2').completer = cli_readerfmt_completer
    subp.add_argument('--uncontract-general', action='store_true', help='Remove general contractions before comparing')


    ########################################
    # Making graphs
    ########################################
    # view-graph
    subp = subparsers.add_parser('view-graph', help='View a file graph for a basis set')
    subp.add_argument('basis', help='Name of the basis set inspect').completer = cli_bsname_completer
    subp.add_argument('--version', help='Which version of the basis set to inspect. Default is the latest version')

    # make-graph-file
    subp = subparsers.add_parser('make-graph-file', help='Make a dot file (and png file) ofr a basis set file graph')
    subp.add_argument('basis', help='Name of the basis set inspect').completer = cli_bsname_completer
    subp.add_argument('outfile', help='Output DOT file to create')
    subp.add_argument('--render', action='store_true', help='Render the DOT file into a corresponding png file')
    subp.add_argument('--version', help='Which version of the basis set to inspect. Default is the latest version')


    #############################
    # DONE WITH SUBCOMMANDS
    #############################

    # setup autocomplete
    argcomplete.autocomplete(parser, validator=cli_case_insensitive_validator)

    # Now parse and handle the args
    args = parser.parse_args()

    # Check and make sure basis sets, roles, etc, are valid
    args = cli_check_normalize_args(args)

    # Actually generate the output
    output = bsecurate_cli_handle_subcmd(args)

    if args.output:
        with open(args.output, 'w') as outfile:
            outfile.write(output)
    else:
        print(output)

    return 0



from ..skel import create_skel


def read_dalton(basis_lines, fname):
    '''Reads Dalton-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the nwchem format does not store all the fields we
       have, so some fields are left blank
    '''

    skipchars = '$'
    basis_lines = [l for l in basis_lines if l and not l[0] in skipchars]

    bs_data = create_skel('component')

    i = 0

    while i < len(basis_lines):
        line = basis_lines[i]

        if line.lower().startswith('a '):
            element_Z = line.split()[1]
            i += 1

            # Shell am is strictly increasing (I hope)
            shell_am = 0

            while i < len(basis_lines) and not basis_lines[i].lower().startswith('a '):
                line = basis_lines[i]
                nprim, ngen = line.split()

                if not element_Z in bs_data['basis_set_elements']:
                    bs_data['basis_set_elements'][element_Z] = {}
                if not 'element_electron_shells' in bs_data['basis_set_elements'][element_Z]:
                    bs_data['basis_set_elements'][element_Z]['element_electron_shells'] = []

                element_data = bs_data['basis_set_elements'][element_Z]

                shell = {
                    'shell_function_type': 'gto',
                    'shell_harmonic_type': 'spherical',
                    'shell_region': '',
                    'shell_angular_momentum': [shell_am]
                }

                exponents = []
                coefficients = []

                i += 1
                for _ in range(int(nprim)):
                    line = basis_lines[i].replace('D', 'E')
                    line = line.replace('d', 'E')
                    lsplt = line.split()
                    exponents.append(lsplt[0])
                    coefficients.append(lsplt[1:])
                    i += 1

                shell['shell_exponents'] = exponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                shell['shell_coefficients'] = list(map(list, zip(*coefficients)))

                # Make sure the number of general contractions is >0
                # (This error was found in some bad files)
                if int(ngen) <= 0:
                    raise RuntimeError("Number of general contractions is not greater than zero for element " + str(element_Z))

                # Make sure the number of general contractions match the heading line
                if len(shell['shell_coefficients']) != int(ngen):
                    raise RuntimeError("Number of general contractions does not equal what was given for element " + str(element_Z))

                element_data['element_electron_shells'].append(shell)
                shell_am += 1

    return bs_data



"""
Tests for creating bundles/archives of formatted data
"""

import os
import zipfile
import tarfile
import pytest
import shutil
import glob

import basis_set_exchange as bse
from .common_testvars import fake_data_dir

_bundle_types = bse.bundle.get_archive_types()
_bundle_exts = [v['extension'] for v in _bundle_types.values()]

def _extract_all(filepath, extract_dir):
    if filepath.endswith('.zip'):
        with zipfile.ZipFile(filepath, 'r') as zf:
            zf.extractall(extract_dir)
    elif filepath.endswith('.tar.bz2'):
        with tarfile.open(filepath, 'r:bz2') as tf:
            tf.extractall(extract_dir)
    else:
        raise RuntimeError("Unexpected file extension")


# yapf: disable
@pytest.mark.slow
@pytest.mark.parametrize('ext', _bundle_exts)
@pytest.mark.parametrize('fmt, reffmt', [('nwchem', 'bib'),
                                         ('psi4', 'txt')])
# yapf: enable
def _run_test_bundles(tmp_path, fmt, reffmt, ext, data_dir):
    '''Test functionality related to creating archive of basis set'''

    tmp_path = str(tmp_path)  # Needed for python 3.5

    bs_ext = bse.converters.get_format_extension(fmt)
    ref_ext = bse.refconverters.get_format_extension(reffmt)

    filename = "bundletest_" + fmt + "_" + reffmt + ext
    filepath = os.path.join(tmp_path, filename)

    bse.create_bundle(filepath, fmt, reffmt, None, data_dir)
    extract_dir = "extract_" + filename
    extract_path = os.path.join(tmp_path, extract_dir)
    _extract_all(filepath, extract_path)

    # Keep track of all the basis sets we have found
    # Start with all found in the data dir, and remove
    # each time we process one
    all_bs = []
    for k,v in bse.get_metadata(data_dir).items():
        for ver in v['versions'].keys():
            all_bs.append((k,ver))

    all_ref = all_bs.copy()

    for root, dirs, files in os.walk(extract_path):
        for basename in files:
            fpath = os.path.join(root, basename)
            name,ver = basename.split('.')[:2]
            if name == 'README':
                continue
            if basename.endswith('.ref' + ref_ext):
                compare_data = bse.get_references(name, fmt=reffmt, version=ver, data_dir=data_dir)
                all_ref.remove((name,ver))
            elif basename.endswith(bs_ext):
                compare_data = bse.get_basis(name, fmt=fmt, version=ver, data_dir=data_dir)
                all_bs.remove((name,ver))
            elif basename.endswith('.family_notes'):
                compare_data = bse.get_family_notes(name, data_dir)
            elif basename.endswith('.notes'):
                compare_data = bse.get_basis_notes(name, data_dir)
            else:
                raise RuntimeError("Unknown file found: " + fpath)

            with open(fpath, 'r') as ftmp:
                assert compare_data == ftmp.read()

    assert len(all_bs) == 0
    assert len(all_ref) == 0


# yapf: disable
@pytest.mark.parametrize('ext', _bundle_exts)
@pytest.mark.parametrize('fmt, reffmt', [('nwchem', 'bib'),
                                         ('psi4', 'txt')])
# yapf: enable
def test_bundles_fast(tmp_path, fmt, reffmt, ext):
   _run_test_bundles(tmp_path, fmt, reffmt, ext, fake_data_dir)


# yapf: disable
@pytest.mark.slow
@pytest.mark.parametrize('ext', _bundle_exts)
@pytest.mark.parametrize('fmt, reffmt', [('nwchem', 'bib'),
                                         ('psi4', 'txt')])
# yapf: enable
def test_bundles_slow(tmp_path, fmt, reffmt, ext):
   _run_test_bundles(tmp_path, fmt, reffmt, ext, None)



'''
Sorting of BSE related dictionaries and data
'''

import sys

# Dictionaries for python 3.6 and above are insertion ordered
# For other versions, use an OrderedDict
_use_odict = sys.version_info.major == 3 and sys.version_info.minor < 6

if _use_odict:
    from collections import OrderedDict


def sort_basis_dict(bs):
    """Sorts a basis set dictionary into a standard order

    This, for example, allows the written file to be more easily read by humans by,
    for example, putting the name and description before more detailed fields.

    This is generally for cosmetic reasons. However, users will generally like things
    in a consistent order
    """

    # yapf: disable
    _keyorder = [
        # Schema stuff
        'molssi_bse_schema', 'schema_type', 'schema_version',

        # Auxiliary block
         'jkfit', 'jfit', 'rifit', 'admmfit', 'dftxfit', 'dftcfit',

        # Basis set metadata
        'basis_set_name', 'basis_set_family', 'basis_set_description', 'basis_set_role', 'basis_set_auxiliaries',
        'basis_set_references', 'basis_set_notes', 'basis_set_function_types',

        # Reference stuff
        'reference_description', 'reference_keys',

        # Version metadata
        'basis_set_version', 'basis_set_revision_description',

        # Elements and data
        'basis_set_elements', 'element_references', 'element_ecp_electrons',
        'element_electron_shells', 'element_ecp', 'element_components', 'element_entry',

        # Shell information
        'shell_function_type', 'shell_harmonic_type', 'shell_region', 'shell_angular_momentum', 'shell_exponents',
        'shell_coefficients',
        'potential_ecp_type', 'potential_angular_momentum', 'potential_r_exponents', 'potential_gaussian_exponents',
        'potential_coefficients'
    ]
    # yapf: enable

    # Add integers for the elements (being optimistic that element 150 will be found someday)
    _keyorder.extend([str(x) for x in range(150)])

    bs_sorted = sorted(bs.items(), key=lambda x: _keyorder.index(x[0]))
    if _use_odict:
        bs_sorted = OrderedDict(bs_sorted)
    else:
        bs_sorted = dict(bs_sorted)

    for k, v in bs_sorted.items():
        # If this is a dictionary, sort recursively
        # If this is a list, sort each element but DO NOT sort the list itself.
        if isinstance(v, dict):
            bs_sorted[k] = sort_basis_dict(v)
        elif isinstance(v, list):
            # Note - the only nested list is with coeffs, which shouldn't be sorted
            #        (so we don't have to recurse into lists of lists)
            bs_sorted[k] = [sort_basis_dict(x) if isinstance(x, dict) else x for x in v]

    return bs_sorted


def sort_single_reference(ref_entry):
    """Sorts a dictionary containing data for a single reference into a standard order
    """

    # yapf: disable
    _keyorder = [
        # Schema stuff
        # This function gets called on the schema 'entry', too
        'schema_type', 'schema_version',

        # Type of the entry
        'type',

        # Actual publication info
        'authors', 'title', 'booktitle', 'series', 'editors', 'journal',
        'institution', 'volume', 'number', 'page', 'year', 'note', 'publisher',
        'address', 'isbn', 'doi'
    ]
    # yapf: enable

    sorted_entry = sorted(ref_entry.items(), key=lambda x: _keyorder.index(x[0]))

    if _use_odict:
        return OrderedDict(sorted_entry)
    else:
        return dict(sorted_entry)


def sort_references_dict(refs):
    """Sorts a reference dictionary into a standard order

    The keys of the references are also sorted, and the keys for the data for each
    reference are put in a more canonical order.
    """

    if _use_odict:
        refs_sorted = OrderedDict()
    else:
        refs_sorted = dict()

    # We insert this first, That is ok - it will be overwritten
    # with the sorted version later
    refs_sorted['molssi_bse_schema'] = refs['molssi_bse_schema']

    # This sorts the entries by reference key (author1985a, etc)
    for k, v in sorted(refs.items()):
        refs_sorted[k] = sort_single_reference(v)

    return refs_sorted



"""
Functions related to composing basis sets from individual components
"""

import os
from . import fileio, manip, memo


def _whole_basis_types(basis):
    '''
    Get a list of all the types of features in this basis set.

    '''

    all_types = set()

    for v in basis['basis_set_elements'].values():
        if 'element_electron_shells' in v:
            for sh in v['element_electron_shells']:
                tstr = '{}_{}'.format(sh['shell_harmonic_type'], sh['shell_function_type'])
                all_types.add(tstr)

        if 'element_ecp' in v:
            for pot in v['element_ecp']:
                all_types.add(pot['potential_ecp_type'] + '_ecp')

    return sorted(list(all_types))


def compose_elemental_basis(file_relpath, data_dir):
    """
    Creates an 'elemental' basis from an elemental json file

    This function reads the info from the given file, and reads all the component
    basis set information from the files listed therein. It then composes all the
    information together into one 'elemental' basis dictionary
    """

    # Do a simple read of the json
    el_bs = fileio.read_json_basis(os.path.join(data_dir, file_relpath))

    # construct a list of all files to read
    component_files = set()
    for k, v in el_bs['basis_set_elements'].items():
        component_files.update(set(v['element_components']))

    # Read all the data from these files into a big dictionary
    component_map = {k: fileio.read_json_basis(os.path.join(data_dir, k)) for k in component_files}

    # Broadcast the basis_set_references to each element
    # Use the basis_set_description for the reference description
    for k, v in component_map.items():
        for el, el_data in v['basis_set_elements'].items():
            el_data['element_references'] = [{
                'reference_description': v['basis_set_description'],
                'reference_keys': v['basis_set_references']
            }]

    # Compose on a per-element basis
    for k, v in el_bs['basis_set_elements'].items():

        components = v.pop('element_components')

        # all of the component data for this element
        el_comp_data = [component_map[c]['basis_set_elements'][k] for c in components]

        # merge all the data
        v = manip.merge_element_data(v, el_comp_data)
        el_bs['basis_set_elements'][k] = v

    return el_bs


@memo.BSEMemoize
def compose_table_basis(file_relpath, data_dir):
    """
    Creates a 'table' basis from an table json file

    This function reads the info from the given file, and reads all the elemental
    basis set information from the files listed therein. It then composes all the
    information together into one 'table' basis dictionary
    """

    # Do a simple read of the json
    file_path = os.path.join(data_dir, file_relpath)
    table_bs = fileio.read_json_basis(file_path)

    # construct a list of all elemental files to read
    element_files = set()
    for v in table_bs['basis_set_elements'].values():
        element_files.add(v['element_entry'])

    # Create a map of the elemental basis data
    # (maps file path to data contained in that file)
    element_map = {k: compose_elemental_basis(k, data_dir) for k in element_files}

    # Replace the basis set for all elements in the table basis with the data
    # from the elemental basis
    for k, v in table_bs['basis_set_elements'].items():
        entry = v['element_entry']
        data = element_map[entry]

        table_bs['basis_set_elements'][k] = data['basis_set_elements'][k]

    # Add the version to the dictionary
    file_base = os.path.basename(file_relpath)
    table_bs['basis_set_version'] = file_base.split('.')[-3]

    # Add whether the entire basis is spherical or cartesian
    table_bs['basis_set_function_types'] = _whole_basis_types(table_bs)

    # Read and merge in the metadata
    # This file must be in the same location as the table file
    meta_dirpath, table_filename = os.path.split(file_path)
    meta_filename = table_filename.split('.')[0] + '.metadata.json'
    meta_filepath = os.path.join(meta_dirpath, meta_filename)
    bs_meta = fileio.read_json_basis(meta_filepath)
    table_bs.update(bs_meta)

    # Remove the molssi schema (which isn't needed here)
    table_bs.pop('molssi_bse_schema')

    return table_bs



'''
Helpers for handling BSE metadata
'''

import os
from collections import OrderedDict

from ..compose import compose_table_basis
from ..fileio import get_all_filelist, _write_plain_json
from ..misc import transform_basis_name


def create_metadata_file(output_path, data_dir):
    '''Creates a METADATA.json file from a data directory

    The file is written to output_path
    '''

    basis_filelist = get_all_filelist(data_dir)[1]

    metadata = {}
    for bs_file_relpath in basis_filelist:
        # Path to the table file (relative to data_dir)
        # metadata files and table files for all versions should be in the
        # same location
        # Also obtain the filename of the table file
        table_relpath, table_filename = os.path.split(bs_file_relpath)

        # Obtain the base filename and version from the filename
        # The base filename is the part before the first period
        # (filebase.ver.table.json)
        table_filebase, ver, _, _ = table_filename.split('.')

        # Fully compose the basis set from components
        bs = compose_table_basis(bs_file_relpath, data_dir)

        # Prepare the metadata
        tr_name = transform_basis_name(bs['basis_set_name'])
        defined_elements = sorted(list(bs['basis_set_elements'].keys()), key=lambda x: int(x))

        # Determine the types of functions contained in the basis
        # (gto, ecp, etc)
        function_types = bs['basis_set_function_types']

        # Create the metadata for this specific version
        # yapf: disable
        version_meta = OrderedDict([('file_relpath', bs_file_relpath),
                                    ('revdesc', bs['basis_set_revision_description']),
                                    ('elements', defined_elements)])
        # yapf: enable

        # Add to the full metadata dict
        if not tr_name in metadata:
            # Create a new entry, with all the common metadata at the top
            # for this entry
            # yapf: disable
            metadata[tr_name] = OrderedDict([
                                 ('display_name', bs['basis_set_name']),
                                 ('description', bs['basis_set_description']),
                                 ('latest_version', None),
                                 ('basename', table_filebase),
                                 ('relpath', table_relpath),
                                 ('family', bs['basis_set_family']),
                                 ('role', bs['basis_set_role']),
                                 ('functiontypes', function_types),
                                 ('auxiliaries', bs['basis_set_auxiliaries']),
                                 ('versions', {ver: version_meta})])
            # yapf: enable

        else:
            # There is an existing entry. Make sure
            # file basename, filename, and functiontypes match
            if metadata[tr_name]['functiontypes'] != function_types:
                print(metadata[tr_name]['functiontypes'], function_types)
                raise RuntimeError("Function types do not match across versions for " + tr_name)
            if metadata[tr_name]['basename'] != table_filebase:
                raise RuntimeError("File basenames do not match across versions for " + tr_name)
            if metadata[tr_name]['relpath'] != table_relpath:
                raise RuntimeError("Basis directories do not match across versions for " + tr_name)

            metadata[tr_name]['versions'][ver] = version_meta

    # sort the versions and find the max version
    for k, v in metadata.items():
        latest = max(v['versions'].keys())
        metadata[k]['latest_version'] = latest

        # Reorder the versions (may have been read out of order)
        metadata[k]['versions'] = OrderedDict(sorted(v['versions'].items()))

    # Write out the metadata
    metadata = OrderedDict(sorted(list(metadata.items())))
    _write_plain_json(output_path, metadata)



#!/usr/bin/env python3

# This script determines what basis sets are missing authoritative sources
# that should be placed in this directory

import os
from basis_set_exchange import get_metadata

my_dir = os.path.dirname(os.path.abspath(__file__))
auth_sources = [x for x in os.listdir(my_dir) if x.endswith('.bz2')]

md = get_metadata()

found_ver = {}

# Start with everything in the metadata, removing what we find
# in this directory
missing = {k:list(v['versions'].keys()) for k,v in md.items()}

for s in auth_sources:
    bsname,ver,_,_ = s.split('.')

    if not bsname in missing:
        raise RuntimeError("Source {} does not correspond to a basis set in the library".format(bsname))
    if not ver in missing[bsname]:
        raise RuntimeError("Source {} version {} does not correspond to a basis set version in the library".format(bsname, ver))

    missing[bsname].remove(ver) 


missing = {k:v for k,v in missing.items() if len(v) > 0}

maxlen = max([len(k) for k in missing.keys()]) + 5

for k,v in missing.items():
    print('{:{maxlen}} {}'.format(k, ','.join(v), maxlen=maxlen))



'''
Miscellaneous curation functions
'''

from .. import fileio
from .. import misc


def elements_in_files(filelist):
    '''Get a list of what elements exist in JSON files

    This works on table, element, and component data files

    Parameters
    ----------
    filelist : list
        A list of paths to json files

    Returns
    -------
    dict
        Keys are the file path, value is a compacted element string of
        what elements are in that file 
    '''

    ret = {}
    for fpath in filelist:
        filedata = fileio.read_json_basis(fpath)
        els = list(filedata['basis_set_elements'].keys())
        ret[fpath] = misc.compact_elements(els)

    return ret



'''
Add a basis set to the library
'''

import os
from ..fileio import write_json_basis
from ..misc import expand_elements
from ..validator import validate_data
from .skel import create_skel
from .readers import read_formatted_basis
from .metadata import create_metadata_file


def add_basis(bs_file,
              data_dir,
              subdir,
              file_base,
              name,
              family,
              role,
              description,
              version,
              revision_description,
              refs=None,
              file_fmt=None):
    '''
    Add a basis set to this library

    This takes in a single file containing the basis set is some format, parses it, and
    create the component, element, and table basis set files in the given data_dir (and subdir).
    The metadata file for the basis is created if it doesn't exist, and the main metadata file is
    also updated.
    
    Parameters
    ----------
    bs_file : str
        Path to the file with formatted basis set information
    data_dir : str
        Path to the data directory to add the data to
    subdir : str
        Subdirectory of the data directory to add the basis set to
    file_base : str
        Base name for new files
    name : str
        Name of the basis set
    family : str
        Family to which this basis set belongs
    role : str
        Role of the basis set (orbital, etc)
    description : str
        Description of the basis set
    version : str
        Version of the basis set
    revision_description : str
        Description of this version of the basis set
    refs : dict or str
        Mapping of references to elements. This can be a dictionary with a compressed
        string of elements as keys and a list of reference strings as values.
        For example, {'H,Li-B,Kr': ['kumar2018a']}

        If a list or string is passed, then those reference(s) will be used for
        all elements.

        Elements that exist in the file but do not have a reference are given the
        usual 'noref' extension and the references entry is empty.
    file_fmt : str
        Format of the input basis data (None = autodetect)
    '''

    # Read the basis set data into a component file, and add the description
    bs_data = read_formatted_basis(bs_file, file_fmt)
    bs_data['basis_set_description'] = description

    if refs is None:
        component_data = [bs_data]
    else:
        # Split out the component data into files based on the reference
        # information. We keep track of which elements we've done so that
        # we can detect dupliates in the references (which would be an error)
        # (and also handle elements with no reference)
        orig_elements = bs_data['basis_set_elements']
        component_data = []
        done_elements = []

        # If a string or list of strings, use that as a reference for all elements
        if isinstance(refs, str):
            bs_data['basis_set_references'] = [refs]
            component_data.append(bs_data)
        elif isinstance(refs, list):
            bs_data['basis_set_references'] = refs
            component_data.append(bs_data)
        else:
            for k, v in refs.items():
                # Expand the string a list of integers (as strings)
                elements = expand_elements(k, True)

                # Make sure we have info for the given elements
                # and that there are no duplicates
                for el in elements:
                    if not el in orig_elements:
                        raise RuntimeError("Element {} not found in file {}".format(el, bs_file))
                    if el in done_elements:
                        raise RuntimeError("Duplicate element {} in reference string {}".format(el, k))

                # Shallow copy, then only include the given elements.
                # Set the ref, and keep track of the done elements
                bs_tmp = bs_data.copy()
                bs_tmp['basis_set_elements'] = {el: y for el, y in orig_elements.items() if el in elements}

                if isinstance(v, str):
                    bs_tmp['basis_set_references'] = [v]
                else:
                    bs_tmp['basis_set_references'] = v

                component_data.append(bs_tmp)
                done_elements.extend(elements)

            # Handle elements without a reference
            noref_elements = set(orig_elements.keys()) - set(done_elements)

            if len(noref_elements) > 0:
                bs_tmp = bs_data.copy()
                bs_tmp['basis_set_elements'] = {el: y for el, y in orig_elements.items() if el in noref_elements}
                bs_tmp['basis_set_references'] = []
                component_data.append(bs_tmp)

    # Start the data files for the element and table json
    element_file_data = create_skel('element')
    element_file_data['basis_set_name'] = name
    element_file_data['basis_set_description'] = description
    element_file_name = '{}.{}.element.json'.format(file_base, version)
    element_file_relpath = os.path.join(subdir, element_file_name)
    element_file_path = os.path.join(data_dir, element_file_relpath)

    table_file_data = create_skel('table')
    table_file_data['basis_set_revision_description'] = revision_description
    table_file_name = '{}.{}.table.json'.format(file_base, version)

    # and the metadata file
    meta_file_data = create_skel('metadata')
    meta_file_data['basis_set_name'] = name
    meta_file_data['basis_set_family'] = family
    meta_file_data['basis_set_description'] = description
    meta_file_data['basis_set_role'] = role
    meta_file_name = '{}.metadata.json'.format(file_base)

    # These get created directly in the top-level data directory
    table_file_path = os.path.join(data_dir, table_file_name)
    meta_file_path = os.path.join(data_dir, meta_file_name)

    # Can just make all the entries for the table file pretty easily
    table_file_entry = {'element_entry': element_file_relpath}
    table_file_data['basis_set_elements'] = {k: table_file_entry for k in bs_data['basis_set_elements'].keys()}

    # Create the filenames for the components
    # Also keep track of where data for each element is (for the element and table files)
    component_file_map = {}
    for cd in component_data:
        if len(cd['basis_set_references']) > 0:
            file_name = file_base + '_' + '_'.join(cd['basis_set_references'])
        else:
            file_name = file_base + '_noref'

        file_name += '.' + str(version) + '.json'
        file_relpath = os.path.join(subdir, file_name)
        file_path = os.path.join(data_dir, file_relpath)
        component_file_map[file_path] = cd

        # Add to the element file data
        # (we add the relative path to the location of the element file,
        # which resides in subdir)
        for el in cd['basis_set_elements'].keys():
            element_file_data['basis_set_elements'][el] = {'element_components': [file_relpath]}

    # Verify all data using the schema
    for file_path, file_data in component_file_map.items():
        validate_data('component', file_data)
    validate_data('element', element_file_data)
    validate_data('table', table_file_data)

    ######################################################################################
    # Before creating any files, check that all the files don't already exist.
    # Yes, there is technically a race condition (files could be created between the
    # check and then actually writing out), but if that happens, you are probably using
    # this library wrong
    #
    # Note that the metadata file may exist already. That is ok
    ######################################################################################
    for file_path in component_file_map.keys():
        if os.path.exists(file_path):
            raise RuntimeError("Component json file {} already exists".format(file_path))
    if os.path.exists(element_file_path):
        raise RuntimeError("Element json file {} already exists".format(element_file_path))
    if os.path.exists(table_file_path):
        raise RuntimeError("Table json file {} already exists".format(table_file_path))

    #############################################
    # Actually create all the files
    #############################################

    # First, create the subdirectory
    subdir_path = os.path.join(data_dir, subdir)
    if not os.path.exists(subdir_path):
        os.makedirs(subdir_path)

    for file_path, file_data in component_file_map.items():
        write_json_basis(file_path, file_data)

    write_json_basis(element_file_path, element_file_data)
    write_json_basis(table_file_path, table_file_data)

    # Create the metadata file if it doesn't exist already
    if not os.path.exists(meta_file_path):
        write_json_basis(meta_file_path, meta_file_data)

    # Update the metadata file
    metadata_file = os.path.join(data_dir, 'METADATA.json')
    create_metadata_file(metadata_file, data_dir)



'''
Completers & validators for argcomplete
'''

import os
from .. import api, curate


def _fix_datadir(data_dir):
    # Handle tilde and variables. These may not have
    # been substituted by the shell at this point
    if data_dir:
        data_dir = os.path.expanduser(data_dir)
        data_dir = os.path.expandvars(data_dir)
    return data_dir


def cli_case_insensitive_validator(s1, s2):
    s1 = s1.lower()
    s2 = s2.lower()
    return s1.startswith(s2)


def cli_bsname_completer(**kwargs):
    # Get the data dir if it has been specified already
    data_dir = _fix_datadir(kwargs['parsed_args'].data_dir)
    return api.get_all_basis_names(data_dir)


def cli_family_completer(**kwargs):
    # Get the data dir if it has been specified already
    data_dir = _fix_datadir(kwargs['parsed_args'].data_dir)
    return api.get_families(data_dir)


def cli_fmt_completer(**kwargs):
    return api.get_formats()


def cli_reffmt_completer(**kwargs):
    return api.get_reference_formats()


def cli_role_completer(**kwargs):
    return api.get_roles()


def cli_readerfmt_completer(**kwargs):
    return curate.get_reader_formats()



'''
Testing of the BSE CLI interface
'''

import os
import subprocess
import pytest

from .common_testvars import fake_data_dir


def _test_cli_cmd(cmd):
    return subprocess.check_output(cmd, shell=True, cwd='/tmp', universal_newlines=True, stderr=subprocess.STDOUT)


bse_cmds = [
    '-V', '-h', '--help', 'list-formats', 'list-ref-formats', 'list-roles', 'get-data-dir', 'list-basis-sets',
    'list-basis-sets -s sto', 'list-basis-sets -f duNNing', 'list-basis-sets -r orbiTAL -f ahlrichs', 'list-families',
    'lookup-by-role def2-tzVp riFit', 'get-basis sto-3g nwchem', 'get-basis cc-pvtz psi4 --elements=1-10',
    'get-basis def2-tzvp turbomole --elements=H-9,11-Ar,cO', 'get-basis cc-pvqz gaussian94 --version=1 --noheader',
    'get-basis 6-31g nwchem --unc-gen --unc-spdf --unc-seg', 'get-basis 6-31g nwchem --opt-gen',
    'get-basis 6-31g nwchem --make-gen', 'get-refs 6-31g txt', 'get-refs sto-3g bib --elements=6-Ne',
    'get-info aug-cc-pv5z', 'get-info def2-tzvp', 'get-notes sto-3g', 'get-family crenbl', 'get-versions cc-pvqz',
    'get-family-notes sto'
]

fakebse_cmds = [
    'list-basis-sets', 'list-basis-sets -s fake', 'list-basis-sets -f bppfake', 'list-families',
    'lookup-by-role bppFAKEbasis jkFit', 'get-basis bppfakebasis nwchem', 'get-refs bppfakebasis txt',
    'get-info bppfakebasis', 'get-family bppfakebasis', 'get-versions bppfakebasis', 'get-family-notes bppfake'
]


@pytest.mark.parametrize('bse_cmd', bse_cmds)
def test_cli(bse_cmd):
    _test_cli_cmd('bse ' + bse_cmd)


@pytest.mark.parametrize('bse_cmd', fakebse_cmds)
def test_cli_datadir(bse_cmd):
    output = _test_cli_cmd('bse -d ' + fake_data_dir + ' ' + bse_cmd)
    assert 'bppfake' in output


@pytest.mark.slow
def test_cli_createbundle(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5
    bfile_path = os.path.join(tmp_path, 'test_bundle.tar.bz2')
    output = _test_cli_cmd('bse create-bundle gaussian94 bib ' + bfile_path)
    assert os.path.isfile(bfile_path)
    assert output.startswith('Created ')


def test_cli_createbundle_datadir(tmp_path):
    tmp_path = str(tmp_path)  # Needed for python 3.5
    bfile_path = os.path.join(tmp_path, 'test_bundle_datadir.tar.bz2')
    output = _test_cli_cmd('bse -d ' + fake_data_dir + ' create-bundle gaussian94 bib ' + bfile_path)
    assert os.path.isfile(bfile_path)
    assert output.startswith('Created ')



'''
Some data common to all tests
'''

import os
import basis_set_exchange as bse

# The directory containing this file
_my_dir = os.path.dirname(os.path.abspath(__file__))

# Use random for getting sets of elements
rand_seed = 39466  # from random.org

# Load all the metadata once
data_dir = bse.api._default_data_dir
bs_metadata = bse.get_metadata()
bs_names = bse.get_all_basis_names()
bs_formats = list(bse.get_formats().keys()) + [None]
ref_formats = list(bse.get_reference_formats().keys()) + [None]
all_families = bse.get_families()
all_roles = bse.get_roles()
true_false = [True, False]

# All basis names/versions combinations
bs_names_vers = []
for k, v in bs_metadata.items():
    for ver in v['versions'].keys():
        bs_names_vers.append((k, ver))

# Directory with some fake data
fake_data_dir = os.path.join(_my_dir, 'fakedata')

# Directory with authoritative sources
auth_data_dir = os.path.join(_my_dir, 'sources')

# Directory with other testing data
test_data_dir = os.path.join(_my_dir, 'test_data')

# All files in the data dir
all_files = bse.fileio.get_all_filelist(data_dir)
all_metadata_files = all_files[0]
all_table_files = all_files[1]
all_element_files = all_files[2]
all_component_files = all_files[3]

# A representative sample of basis sets
bs_names_sample = ['6-31g', '6-31+g*', 'aug-cc-pvtz', 'lanl2dz', 'def2-tzvp', 'tzp', 'sto-3g']


def bool_matrix(size):
    '''Returns an identity matrix of a given size consisting of bool types
    '''
    ret = [[False for i in range(size)] for j in range(size)]
    for x in range(size):
        ret[x][x] = True
    return ret



'''
Converts basis set data to a specified output format
'''

from .. import sort
from .bsejson import write_json
from .nwchem import write_nwchem
from .g94 import write_g94
from .gamess_us import write_gamess_us
from .psi4 import write_psi4
from .turbomole import write_turbomole

_converter_map = {
    'nwchem': {
        'display': 'NWChem',
        'extension': '.nw',
        'comment': '#',
        'valid': set(['cartesian_gto', 'spherical_gto', 'scalar_ecp']),
        'function': write_nwchem
    },
    'gaussian94': {
        'display': 'Gaussian94',
        'extension': '.gbs',
        'comment': '!',
        'valid': set(['cartesian_gto', 'spherical_gto', 'scalar_ecp']),
        'function': write_g94
    },
    'psi4': {
        'display': 'Psi4',
        'extension': '.gbs',
        'comment': '!',
        'valid': set(['cartesian_gto', 'spherical_gto', 'scalar_ecp']),
        'function': write_psi4
    },
    'gamess_us': {
        'display': 'GAMESS US',
        'extension': '.bas',
        'comment': '!',
        'valid': set(['cartesian_gto', 'spherical_gto', 'scalar_ecp']),
        'function': write_gamess_us
    },
    'turbomole': {
        'display': 'Turbomole',
        'extension': '.tm',
        'comment': '#',
        'valid': set(['cartesian_gto', 'spherical_gto', 'scalar_ecp']),
        'function': write_turbomole
    },
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': None,
        'valid': None,
        'function': write_json
    }
}


def convert_basis(basis_dict, fmt, header=None):
    '''
    Returns the basis set data as a string representing
    the data in the specified output format
    '''

    # Sort the basis dictionary
    basis_dict_sorted = sort.sort_basis_dict(basis_dict)

    # make converters case insensitive
    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    converter = _converter_map[fmt]

    # Determine if the converter supports all the types in the basis_dict
    if converter['valid'] is not None:
        ftypes = set(basis_dict['basis_set_function_types'])
        if ftypes > converter['valid']:
            raise RuntimeError('Converter {} does not support all function types: {}'.format(fmt, str(ftypes)))

    # Actually do the conversion
    ret_str = converter['function'](basis_dict_sorted)

    if header is not None and fmt != 'json':
        comment_str = _converter_map[fmt]['comment']
        header_str = comment_str + comment_str.join(header.splitlines(True))
        ret_str = header_str + '\n\n' + ret_str

    # HACK - Psi4 requires the first non-comment line be spherical/cartesian
    #        so we have to add that before the header
    if fmt == 'psi4':
        types = basis_dict['basis_set_function_types']
        harm_type = 'spherical' if 'spherical_gto' in types else 'cartesian'
        ret_str = harm_type + '\n\n' + ret_str

    return ret_str


def get_formats(function_types=None):
    '''
    Returns the available formats mapped to display name.

    This is returned as an ordered dictionary, with the most common
    at the top, followed by the rest in alphabetical order

    If a list is specified for function_types, only those formats
    supporting the given function types will be returned.
    '''

    if function_types is None:
        return {k: v['display'] for k, v in _converter_map.items()}

    ftypes = [x.lower() for x in function_types]
    ftypes = set(ftypes)
    ret = []

    for fmt,v in _converter_map.items():
        if v['valid'] is None or ftypes <= v['valid']:
           ret.append(fmt) 
    return ret


def get_format_extension(fmt):
    '''
    Returns the recommended extension for a given format
    '''

    if fmt is None:
        return 'dict'

    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    return _converter_map[fmt]['extension']



"""
Tests for BSE metadata
"""

import json
import os
import pytest
import glob

from basis_set_exchange import api, curate, fileio
from .common_testvars import data_dir, all_table_files, all_metadata_files


def test_get_metadata():
    '''Test the get_metadata function in the API'''

    api.get_metadata(data_dir)


def test_metadata_uptodate(tmp_path):
    '''Tests that the METADATA.json file is up to date'''

    tmp_path = str(tmp_path)  # Needed for python 3.5
    old_metadata = os.path.join(data_dir, 'METADATA.json')

    # Create a temporary file
    new_metadata = os.path.join(tmp_path, 'NEW_METADATA.json')

    curate.create_metadata_file(str(new_metadata), data_dir)

    with open(old_metadata, 'r') as f:
        old_data = json.load(f)
    with open(new_metadata, 'r') as f:
        new_data = json.load(f)

    assert old_data == new_data


@pytest.mark.parametrize('meta_file_path', all_metadata_files)
def test_basis_metadata_pair1(meta_file_path):
    '''Test that each metadata file is paired with a table basis
       in the same directory
    '''

    bsname = os.path.basename(meta_file_path).split('.')[0]
    meta_subdir = os.path.dirname(meta_file_path)
    test_path = os.path.join(data_dir, meta_subdir, bsname)
    test_files = glob.glob(test_path + '.*.table.json')
    assert len(test_files) > 0


@pytest.mark.parametrize('table_file_path', all_table_files)
def test_basis_metadata_pair2(table_file_path):
    '''Test that each table basis is paired with a metadata file
       in the same directory
    '''

    bsname = os.path.basename(table_file_path).split('.')[0]
    table_subdir = os.path.dirname(table_file_path)
    test_file = os.path.join(data_dir, table_subdir, bsname) + '.metadata.json'
    assert os.path.isfile(test_file)



"""
Tests for the BSE main API
"""

import random
import pytest

import basis_set_exchange as bse

from .common_testvars import *

# Use random for getting sets of elements
random.seed(rand_seed, version=2)

# To test role lookup
# yapf: disable
role_tests = [('cc-pvdz', 'rifit', 'cc-pvdz-rifit'),
              ('def2-tzvp', 'jfit', 'def2-universal-jfit'),
              ('aug-cc-pv5z', 'jkfit', 'cc-pv5z-jkfit'),
              ('aug-cc-pv5z', 'jkfit', 'cc-pv5z-jkfit'),
              ('aug-pcseg-1', 'admmfit', 'aug-admm-1')]
# yapf: enable


@pytest.mark.parametrize('basis_name', bs_names)
def test_get_basis_1(basis_name):
    """For all versions of basis sets, test a simple get_basis
    """
    this_metadata = bs_metadata[basis_name]
    for ver in this_metadata['versions'].keys():
        bse.get_basis(basis_name, version=ver)


@pytest.mark.parametrize('basis_name', bs_names)
def test_get_basis_2(basis_name):
    """For all versions of basis sets, test a simple get_basis
       with different element selections
    """
    this_metadata = bs_metadata[basis_name]
    latest = this_metadata['latest_version']
    avail_elements = this_metadata['versions'][latest]['elements']
    nelements = random.randint(1, len(avail_elements))
    selected_elements = random.sample(avail_elements, nelements)

    # Change some selected elements to strings
    for idx in range(len(selected_elements)):
        if idx % 3 == 1:
            selected_elements[idx] = bse.lut.element_sym_from_Z(selected_elements[idx])
        elif idx % 3 == 2:
            selected_elements[idx] = str(selected_elements[idx])

    bs = bse.get_basis(basis_name, elements=selected_elements)
    assert len(bs['basis_set_elements']) == len(selected_elements)

    # Try to get as an integer
    bs = bse.get_basis(basis_name, elements=int(selected_elements[0]))
    assert len(bs['basis_set_elements']) == 1


@pytest.mark.parametrize('basis_name', bs_names_sample)
@pytest.mark.parametrize('bool_opts', bool_matrix(5))
def test_get_basis_3(basis_name, bool_opts):
    """For a sample of basis sets, test different options
    """
    bse.get_basis(
        basis_name,
        uncontract_general=bool_opts[0],
        uncontract_segmented=bool_opts[1],
        uncontract_spdf=bool_opts[2],
        make_general=bool_opts[3],
        optimize_general=bool_opts[4])


@pytest.mark.parametrize('basis_name', bs_names_sample)
@pytest.mark.parametrize('fmt', bs_formats)
def test_get_basis_4(basis_name, fmt):
    """For a sample of basis sets, test getting different formats
       of the latest version
    """
    bse.get_basis(basis_name, fmt=fmt)


@pytest.mark.parametrize('basis_name', bs_names)
@pytest.mark.parametrize('fmt', ref_formats)
def test_get_references_1(basis_name, fmt):
    """ Tests getting references for all basis sets

    Also test getting references for a random selection of elements
    """
    this_metadata = bs_metadata[basis_name]
    for ver in this_metadata['versions'].keys():
        bse.get_references(basis_name, fmt=fmt, version=ver)

        avail_elements = this_metadata['versions'][ver]['elements']
        nelements = random.randint(1, len(avail_elements))
        selected_elements = random.sample(avail_elements, nelements)
        bse.get_references(basis_name, elements=selected_elements, fmt=fmt, version=ver)


@pytest.mark.parametrize('primary_basis,role,expected', role_tests)
def test_lookup_by_role(primary_basis, role, expected):
    """Test looking up data by role
    """
    bs = bse.lookup_basis_by_role(primary_basis, role)
    assert bs.lower() == expected.lower()


@pytest.mark.parametrize('basis_name', bs_names)
def test_notes(basis_name):
    """Test getting family, family notes, and basis set notes
    """
    bse.get_basis_notes(basis_name)
    bse.has_basis_notes(basis_name)


@pytest.mark.parametrize('basis_name', bs_names)
def test_get_family(basis_name):
    """Test getting family"""

    fam = bse.get_basis_family(basis_name)
    assert fam in all_families


@pytest.mark.parametrize('family', all_families)
def test_family_notes(family):
    """Test getting family notes"""
    bse.has_family_notes(family)
    bse.get_family_notes(family)


# yapf: disable
@pytest.mark.parametrize('substr,family,role', [['def2', 'ahlrichs', 'orbital'],
                                                ['pVDz', None, None],
                                                [None, None, 'jkfit'],
                                                [None, 'pople', None]])
# yapf: enable
def test_filter(substr, family, role):
    """Test filtering basis set"""
    md = bse.filter_basis_sets(substr, family, role)
    assert len(md) > 0


# yapf: disable
@pytest.mark.parametrize('substr,family,role', [['def2', 'ahlrichs', 'jkfit'],
                                                ['qqqqq', None, None],
                                                ['6-31', None, 'admmfit']])
# yapf: enable
def test_filter_0(substr, family, role):
    """Test filtering basis set (returning zero results)"""
    md = bse.filter_basis_sets(substr, family, role)
    assert len(md) == 0

# yapf: disable
@pytest.mark.parametrize('fmts', [None,
                                  ['spherical_gto', 'scalar_ecp'],
                                  ['CARTESIAN_gto']])
# yapf: enable
def test_get_formats(fmts):
    '''Test the get_formats function'''
    ret = bse.get_formats(fmts)

    # JSON is always supported
    assert len(ret) > 1
